# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Editor panel dialogs for device objects (Host, Firewall, Interface)."""

from datetime import UTC, datetime

from PySide6.QtCore import Slot
from PySide6.QtWidgets import QDialog

from firewallfabrik.gui.base_object_dialog import BaseObjectDialog
from firewallfabrik.gui.iptables_settings_dialog import IptablesSettingsDialog
from firewallfabrik.gui.linux_settings_dialog import LinuxSettingsDialog
from firewallfabrik.gui.nftables_settings_dialog import NftablesSettingsDialog
from firewallfabrik.gui.platform_settings import (
    HOST_OS,
    PLATFORMS,
    get_enabled_os,
    get_enabled_platforms,
)

# Reverse mapping: display name → internal key for host OS.
_HOST_OS_INTERNAL = {v: k for k, v in HOST_OS.items()}

# Platform display name → settings dialog class.
_PLATFORM_SETTINGS_DIALOG = {
    'iptables': IptablesSettingsDialog,
    'nftables': NftablesSettingsDialog,
}


class HostDialog(BaseObjectDialog):
    def __init__(self, parent=None):
        super().__init__('hostdialog_q.ui', parent)

    def _populate(self):
        self.obj_name.setText(self._obj.name or '')
        data = self._obj.data or {}
        self.MACmatching.setChecked(data.get('mac_filter_enabled') == 'True')

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        data = dict(self._obj.data or {})
        data['mac_filter_enabled'] = str(self.MACmatching.isChecked())
        self._obj.data = data


class FirewallDialog(BaseObjectDialog):
    def __init__(self, parent=None):
        super().__init__('firewalldialog_q.ui', parent)

    def _populate(self):
        self.obj_name.setText(self._obj.name or '')
        data = self._obj.data or {}

        # Populate combos with enabled entries before setting the current value.
        self.platform.clear()
        for display in get_enabled_platforms().values():
            self.platform.addItem(display)
        self.hostOS.clear()
        for display in get_enabled_os().values():
            self.hostOS.addItem(display)

        self._set_combo_text(self.platform, data.get('platform', ''))
        self._set_combo_text(self.version, data.get('version', ''))
        host_os = data.get('host_OS', '')
        self._set_combo_text(self.hostOS, HOST_OS.get(host_os, host_os))
        self.inactive.setChecked(data.get('inactive') in (True, 'True'))
        for attr, key in (
            ('last_modified', 'lastModified'),
            ('last_compiled', 'lastCompiled'),
            ('last_installed', 'lastInstalled'),
        ):
            ts = int(data.get(key, 0) or 0)
            text = (
                datetime.fromtimestamp(ts, tz=UTC).strftime('%Y-%m-%d %H:%M:%S')
                if ts
                else '-'
            )
            getattr(self, attr).setText(text)

        self.platform.currentTextChanged.connect(self._update_settings_buttons)
        self.hostOS.currentTextChanged.connect(self._update_settings_buttons)
        self._update_settings_buttons()

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        data = dict(self._obj.data or {})
        data['platform'] = self.platform.currentText()
        data['version'] = self.version.currentText()
        host_os_text = self.hostOS.currentText()
        data['host_OS'] = _HOST_OS_INTERNAL.get(host_os_text, host_os_text)
        data['inactive'] = self.inactive.isChecked()
        self._obj.data = data

    def _update_settings_buttons(self):
        self.fwAdvanced.setEnabled(self.platform.currentText() in PLATFORMS.values())
        self.osAdvanced.setEnabled(self.hostOS.currentText() in HOST_OS.values())

    @Slot()
    def openFWDialog(self):
        dialog_cls = _PLATFORM_SETTINGS_DIALOG.get(self.platform.currentText())
        if dialog_cls is None:
            return
        dlg = dialog_cls(self._obj, parent=self.window())
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.changed.emit()

    @Slot()
    def openOSDialog(self):
        dlg = LinuxSettingsDialog(
            self._obj, platform=self.platform.currentText(), parent=self.window()
        )
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.changed.emit()

    @staticmethod
    def _set_combo_text(combo, text):
        idx = combo.findText(text)
        if idx >= 0:
            combo.setCurrentIndex(idx)
        elif text:
            combo.addItem(text)
            combo.setCurrentIndex(combo.count() - 1)


class InterfaceDialog(BaseObjectDialog):
    def __init__(self, parent=None):
        super().__init__('interfacedialog_q.ui', parent)

    def _populate(self):
        self.obj_name.setText(self._obj.name or '')
        data = self._obj.data or {}
        self.label.setText(data.get('label', ''))
        self.seclevel.setValue(int(data.get('security_level', 0)))
        self.management.setChecked(bool(data.get('management', False)))
        self.unprotected.setChecked(bool(data.get('unprotected', False)))
        self.dedicated_failover.setChecked(bool(data.get('dedicated_failover', False)))
        if data.get('dyn', False):
            self.dynamic.setChecked(True)
        elif data.get('unnum', False):
            self.unnumbered.setChecked(True)
        else:
            self.regular.setChecked(True)

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        data = dict(self._obj.data or {})
        data['label'] = self.label.text()
        data['security_level'] = str(self.seclevel.value())
        data['management'] = self.management.isChecked()
        data['unprotected'] = self.unprotected.isChecked()
        data['dedicated_failover'] = self.dedicated_failover.isChecked()
        data['dyn'] = self.dynamic.isChecked()
        data['unnum'] = self.unnumbered.isChecked()
        self._obj.data = data

    @Slot()
    def openIfaceDialog(self):
        pass  # https://github.com/Linuxfabrik/firewallfabrik/issues/13



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Expected output regression tests for the nftables compiler."""

import pytest

from .conftest import EXPECTED_OUTPUT_DIR, _find_fixture, discover_test_cases
from .normalize import normalize_nft

_CASES = discover_test_cases('nft')


@pytest.mark.parametrize(
    ('fixture_name', 'fw_name'),
    _CASES,
    ids=[f'{f}/{n}' for f, n in _CASES],
)
def test_nftables_expected_output(fixture_name, fw_name, compile_nft, tmp_path):
    """Compile fixture and compare normalized output to expected output."""
    fixture_path = _find_fixture(fixture_name)
    expected_path = EXPECTED_OUTPUT_DIR / 'nft' / fixture_name / f'{fw_name}.fw'

    output_path = compile_nft(fixture_path, fw_name, tmp_path)

    actual = normalize_nft(output_path.read_text())
    expected = expected_path.read_text()

    # Save normalized output next to raw output for easier diffing
    normalized_path = output_path.with_suffix('.normalized.fw')
    normalized_path.write_text(actual)

    assert actual == expected, (
        f'nftables output differs from expected output.\n'
        f'  Actual (raw):        {output_path}\n'
        f'  Actual (normalized): {normalized_path}\n'
        f'  Expected:            {expected_path}\n'
        f'Run "python tests/update_expected_output.py --fixture {fixture_name} --platform nft" to update.'
    )



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""QTreeView wrapper for policy rule display with group and editing support."""

import json
import uuid

from PySide6.QtCore import (
    QEvent,
    QItemSelectionModel,
    QMimeData,
    QModelIndex,
    QRect,
    QSettings,
    QSize,
    Qt,
)
from PySide6.QtGui import QColor, QDrag, QFont, QIcon, QPalette
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QHBoxLayout,
    QHeaderView,
    QInputDialog,
    QLabel,
    QMenu,
    QStyle,
    QStyledItemDelegate,
    QToolButton,
    QToolTip,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from firewallfabrik.gui.policy_context_menu import (
    VALID_TYPES_BY_SLOT,
    add_color_submenu,
    add_new_group_action,
    add_to_adjacent_group_actions,
    build_action_menu,
    build_comment_menu,
    build_direction_menu,
    build_element_menu,
    build_group_header_menu,
    build_metric_menu,
    build_options_menu,
    build_row_menu,
)
from firewallfabrik.gui.policy_model import (
    ELEMENTS_ROLE,
    FWF_MIME_TYPE,
    NEGATED_ROLE,
)


class _CellBorderDelegate(QStyledItemDelegate):
    """Delegate that draws lightgray cell borders, vertical padding, and
    renders element columns with per-object icons stacked vertically.

    Matches fwbuilder's ``RuleSetViewDelegate`` look.
    """

    _BORDER_COLOR = QColor('lightgray')
    _H_PAD = 2
    _HIGHLIGHT_COLOR = QColor(255, 255, 150, 100)
    _ICON_TEXT_GAP = 2
    _V_PAD = 2

    def _icon_size(self):
        """Return the configured icon size (16 or 25)."""
        return QSettings().value('UI/IconSizeInRules', 25, type=int)

    def _icon_suffix(self, negated=False):
        """Return the QRC alias suffix for the configured size."""
        if self._icon_size() == 16:
            return 'icon-neg-tree' if negated else 'icon-tree'
        return 'icon-neg' if negated else 'icon'

    def sizeHint(self, option, index):
        icon_sz = self._icon_size()
        fm = option.fontMetrics
        line_h = max(icon_sz, fm.height())
        elements = index.data(ELEMENTS_ROLE)
        if elements:
            height = line_h * len(elements) + 2 * self._V_PAD
            max_text_w = max(fm.horizontalAdvance(name) for _, name, *_ in elements)
            width = icon_sz + self._ICON_TEXT_GAP + max_text_w + 2 * self._H_PAD
            return QSize(width, height)
        return QSize(super().sizeHint(option, index).width(), line_h + 2 * self._V_PAD)

    def paint(self, painter, option, index):
        # Manual background painting matching fwbuilder's RuleSetViewDelegate:
        # flat fillRect instead of styled CE_ItemViewItem to avoid platform
        # selection effects (gradients, rounded corners, etc.).
        bg = index.data(Qt.ItemDataRole.BackgroundRole)
        if bg:
            painter.fillRect(option.rect, bg)
        elif option.state & QStyle.StateFlag.State_Selected:
            painter.fillRect(
                option.rect,
                option.palette.color(QPalette.ColorRole.Highlight),
            )

        # Always highlight the position column for the current row (like
        # Excel row headers) so the user can orient themselves.
        view = self.parent()
        model = index.model()
        if (
            isinstance(view, PolicyView)
            and model is not None
            and index.column() == model.position_col
            and not (option.state & QStyle.StateFlag.State_Selected)
        ):
            cur = view.currentIndex()
            if (
                cur.isValid()
                and cur.row() == index.row()
                and cur.parent() == index.parent()
            ):
                painter.fillRect(
                    option.rect,
                    option.palette.color(QPalette.ColorRole.Highlight),
                )

        # Highlight matched cell (Find / Where Used navigation).
        view = self.parent()
        if (
            isinstance(view, PolicyView)
            and view._highlight_rule_id is not None
            and view._highlight_col is not None
            and index.column() == view._highlight_col
        ):
            rd = index.model().get_row_data(index)
            if rd is not None and rd.rule_id == view._highlight_rule_id:
                painter.fillRect(option.rect, self._HIGHLIGHT_COLOR)

        elements = index.data(ELEMENTS_ROLE)
        if elements:
            self._paint_elements(painter, option, index, elements)
        else:
            self._paint_cell(painter, option, index)

        # Cell border.
        painter.save()
        painter.setPen(self._BORDER_COLOR)
        painter.drawRect(option.rect.adjusted(0, 0, -1, -1))
        painter.restore()

    def _paint_cell(self, painter, option, index):
        """Paint a single-value cell (icon + text) top-aligned."""
        # Per-element highlight for non-element columns (Action, Direction, Options).
        cell_highlighted = False
        view = self.parent()
        if hasattr(view, '_selected_element') and view._selected_element is not None:
            sel_index = view._selected_index
            if (
                index.row() == sel_index.row()
                and index.column() == sel_index.column()
                and index.parent() == sel_index.parent()
            ):
                if view.hasFocus():
                    painter.fillRect(
                        option.rect,
                        option.palette.color(QPalette.ColorRole.Highlight),
                    )
                    cell_highlighted = True
                else:
                    painter.save()
                    painter.setPen(QColor('red'))
                    painter.drawRect(option.rect.adjusted(0, 0, -1, -1))
                    painter.restore()

        icon_sz = self._icon_size()
        rect = option.rect.adjusted(self._H_PAD, self._V_PAD, -self._H_PAD, 0)
        line_h = max(icon_sz, painter.fontMetrics().height())
        fg = index.data(Qt.ItemDataRole.ForegroundRole)
        alignment = index.data(Qt.ItemDataRole.TextAlignmentRole)
        h_align = Qt.AlignmentFlag.AlignLeft
        if alignment and alignment & Qt.AlignmentFlag.AlignRight:
            h_align = Qt.AlignmentFlag.AlignRight

        painter.save()
        if cell_highlighted:
            painter.setPen(
                option.palette.color(QPalette.ColorRole.HighlightedText),
            )
        elif fg:
            painter.setPen(fg.color() if hasattr(fg, 'color') else fg)

        x = rect.left()
        icon = index.data(Qt.ItemDataRole.DecorationRole)
        if isinstance(icon, QIcon) and not icon.isNull():
            icon.paint(painter, QRect(x, rect.top(), icon_sz, line_h))
            x += icon_sz + self._ICON_TEXT_GAP

        text = index.data(Qt.ItemDataRole.DisplayRole)
        if text is not None and text != '':
            text_rect = QRect(x, rect.top(), rect.right() - x, line_h)
            painter.drawText(
                text_rect,
                h_align | Qt.AlignmentFlag.AlignVCenter,
                str(text),
            )
        painter.restore()

    def _paint_elements(self, painter, option, index, elements):
        """Paint a list of (id, name, type) elements with icons."""
        icon_sz = self._icon_size()
        negated = index.data(NEGATED_ROLE)
        icon_suffix = self._icon_suffix(negated=negated)
        rect = option.rect.adjusted(self._H_PAD, self._V_PAD, -self._H_PAD, 0)
        line_h = max(icon_sz, painter.fontMetrics().height())
        fg = index.data(Qt.ItemDataRole.ForegroundRole)

        # Per-element highlight from the view's selection.
        view = self.parent()
        sel_target_id = None
        if hasattr(view, '_selected_element') and view._selected_element is not None:
            sel_index = view._selected_index
            if (
                index.row() == sel_index.row()
                and index.column() == sel_index.column()
                and index.parent() == sel_index.parent()
            ):
                sel_target_id = view._selected_element[2]
        has_focus = view.hasFocus() if sel_target_id is not None else False

        painter.save()
        if fg:
            painter.setPen(fg.color() if hasattr(fg, 'color') else fg)
        default_pen = painter.pen()
        for target_id, name, obj_type, *_ in elements:
            # Draw per-element selection highlight.
            if sel_target_id is not None and target_id == sel_target_id:
                elem_rect = QRect(rect.left(), rect.top(), rect.width(), line_h)
                if has_focus:
                    painter.fillRect(
                        elem_rect,
                        option.palette.color(QPalette.ColorRole.Highlight),
                    )
                    painter.setPen(
                        option.palette.color(QPalette.ColorRole.HighlightedText),
                    )
                else:
                    painter.save()
                    painter.setPen(QColor('red'))
                    painter.drawRect(elem_rect.adjusted(0, 0, -1, -1))
                    painter.restore()

            icon_path = f':/Icons/{obj_type}/{icon_suffix}' if obj_type else ''
            x = rect.left()
            if icon_path:
                icon = QIcon(icon_path)
                if not icon.isNull():
                    icon.paint(
                        painter,
                        QRect(x, rect.top(), icon_sz, line_h),
                    )
                x += icon_sz + self._ICON_TEXT_GAP
            text_rect = QRect(x, rect.top(), rect.right() - x, line_h)
            painter.drawText(
                text_rect,
                Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
                name,
            )
            # Restore default pen after highlighted element.
            if sel_target_id is not None and target_id == sel_target_id and has_focus:
                painter.setPen(default_pen)
            rect.setTop(rect.top() + line_h)
        painter.restore()


def _model_col_to_slot(model, col):
    """Return the slot name for *col* using model config."""
    return model.col_to_slot.get(col)


def _model_element_cols(model):
    """Return the element column frozenset from the model."""
    return model.element_cols


def _model_selectable_cols(model):
    """Return the selectable column frozenset from the model."""
    return model.selectable_cols


class PolicyView(QTreeView):
    """Tree view with context menus, keyboard shortcuts, and drop support."""

    def __init__(self, clipboard_store, bridge=None, parent=None):
        super().__init__(parent)
        self._clipboard_store = clipboard_store
        self._bridge = bridge
        self._highlight_rule_id = None
        self._highlight_col = None
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectItems)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setRootIsDecorated(True)
        self.setItemDelegate(_CellBorderDelegate(self))
        self._apply_icon_size()
        self.header().setStretchLastSection(True)
        self.header().setSectionResizeMode(
            QHeaderView.ResizeMode.Interactive,
        )
        self.header().sectionHandleDoubleClicked.connect(
            self.resizeColumnToContents,
        )

        # Context menu (handled via contextMenuEvent override).

        # Double-click on group header -> rename.
        self.doubleClicked.connect(self._on_double_clicked)

        # Drop support.
        self.setAcceptDrops(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DropOnly)
        self.setDropIndicatorShown(True)

        # Per-element selection state.
        self._selected_element = None  # (rule_id, slot, target_id) or None
        self._selected_index = QModelIndex()  # cell containing selected element
        self._drag_start_pos = None  # QPoint for drag threshold

    def set_highlight(self, rule_id, col):
        """Mark a single cell for visual emphasis."""
        self._highlight_rule_id = rule_id
        self._highlight_col = col
        self.viewport().update()

    def clear_highlight(self):
        """Remove cell highlight."""
        if self._highlight_rule_id is not None:
            self._highlight_rule_id = None
            self._highlight_col = None
            self.viewport().update()

    def viewportEvent(self, event):
        """Show per-element tooltips and an instructional tooltip for empty space."""
        if event.type() == QEvent.Type.ToolTip:
            if not QSettings().value('UI/ObjTooltips', True, type=bool):
                return super().viewportEvent(event)
            index = self.indexAt(event.pos())
            if not index.isValid():
                QToolTip.showText(
                    event.globalPos(),
                    '<html>'
                    'Policy, NAT and routing rules are shown here.'
                    '<ul>'
                    '<li><b>Rules use objects</b> &ndash; to use an object '
                    'like an IP address in a rule, first create it in the '
                    'object tree.</li>'
                    '<li><b>Drag and drop</b> objects from the tree to the '
                    'desired field (Source, Destination, etc.) in the rule.</li>'
                    '<li><b>To add a rule</b>, click the &ldquo;+&rdquo; '
                    'button at the top of the window.</li>'
                    '<li><b>To open the context menu</b> with operations such '
                    "as 'Add rule', 'Remove rule', etc., right-click.</li>"
                    '</ul></html>',
                    self,
                )
                return True
            # Per-element tooltip: show only the hovered element's tooltip.
            elements = index.data(ELEMENTS_ROLE)
            if elements and len(elements) > 1:
                tip = self._element_tooltip_at(index, event.pos(), elements)
                if tip:
                    QToolTip.showText(event.globalPos(), tip, self)
                    return True
        return super().viewportEvent(event)

    def _element_tooltip_at(self, index, pos, elements):
        """Return the tooltip for the element at *pos*, or None."""
        rect = self.visualRect(index)
        icon_sz = QSettings().value('UI/IconSizeInRules', 25, type=int)
        fm = self.fontMetrics()
        line_h = max(icon_sz, fm.height())
        v_pad = _CellBorderDelegate._V_PAD
        y_offset = pos.y() - rect.top() - v_pad
        if y_offset < 0:
            return None
        elem_idx = y_offset // line_h
        if 0 <= elem_idx < len(elements):
            *_, tip = elements[elem_idx]
            return tip or None
        return None

    # ------------------------------------------------------------------
    # Model setup
    # ------------------------------------------------------------------

    def setModel(self, model):
        super().setModel(model)
        if model is not None:
            model.modelAboutToBeReset.connect(self._save_selection)
            model.modelReset.connect(self._configure_groups)
            self._configure_groups()

    def _apply_icon_size(self):
        """Set the view's icon size from the user preference."""
        sz = QSettings().value('UI/IconSizeInRules', 25, type=int)
        self.setIconSize(QSize(sz, sz))

    def _configure_groups(self):
        """Mark group rows as first-column-spanned, expand, fit columns, and restore selection."""
        model = self.model()
        if model is None:
            return
        self._apply_icon_size()
        for row in range(model.rowCount(QModelIndex())):
            idx = model.index(row, 0, QModelIndex())
            if model.is_group(idx):
                self.setFirstColumnSpanned(row, QModelIndex(), True)
        self.expandAll()
        for col in range(model.columnCount()):
            self.resizeColumnToContents(col)
        self._restore_selection()

    # ------------------------------------------------------------------
    # Selection save / restore across model resets
    # ------------------------------------------------------------------

    def _save_selection(self):
        """Snapshot row and element selections before a model reset."""
        model = self.model()
        self._saved_row_rule_ids = []
        self._saved_current_rule_id = None
        if model is not None:
            for idx in self._selected_rule_indices():
                rd = model.get_row_data(idx)
                if rd is not None:
                    self._saved_row_rule_ids.append(rd.rule_id)
            cur = self.currentIndex()
            if cur.isValid():
                rd = model.get_row_data(cur)
                if rd is not None:
                    self._saved_current_rule_id = rd.rule_id
        self._saved_element = self._selected_element  # (rule_id, slot, target_id)
        # Save the element's position within its cell so we can fall
        # back to the next element if the saved one was deleted.
        self._saved_element_pos = None
        if self._selected_element is not None and self._selected_index.isValid():
            elements = self._selected_index.data(ELEMENTS_ROLE) or []
            target_id = self._selected_element[2]
            for i, (eid, *_rest) in enumerate(elements):
                if eid == target_id:
                    self._saved_element_pos = i
                    break
        # Clear stale index references (they become invalid after reset).
        self._selected_element = None
        self._selected_index = QModelIndex()
        self._drag_start_pos = None

    def _restore_selection(self):
        """Re-select rows and the per-element highlight after a model reset."""
        model = self.model()
        if model is None:
            return

        sel_model = self.selectionModel()
        slot_to_col = model.slot_to_col

        # Restore current index (needed for keyboard shortcuts like X to compile).
        saved_current = getattr(self, '_saved_current_rule_id', None)
        if saved_current is not None:
            idx = model.index_for_rule(saved_current)
            if idx.isValid():
                sel_model.setCurrentIndex(
                    idx,
                    QItemSelectionModel.SelectionFlag.NoUpdate,
                )
            self._saved_current_rule_id = None

        # Restore row selection.
        saved_ids = getattr(self, '_saved_row_rule_ids', [])
        if saved_ids:
            for rule_id in saved_ids:
                idx = model.index_for_rule(rule_id)
                if idx.isValid():
                    sel_model.select(
                        idx,
                        QItemSelectionModel.SelectionFlag.Select,
                    )
            self._saved_row_rule_ids = []

        # Restore per-element selection.
        saved_elem = getattr(self, '_saved_element', None)
        if saved_elem is not None:
            rule_id, slot, target_id = saved_elem
            idx = model.index_for_rule(rule_id)
            if idx.isValid():
                col = slot_to_col.get(slot)
                if col is not None:
                    cell_idx = model.index(idx.row(), col, idx.parent())
                    elements = cell_idx.data(ELEMENTS_ROLE) or []
                    # Check if the saved element still exists.
                    if any(eid == target_id for eid, *_ in elements):
                        self._selected_element = saved_elem
                        self._selected_index = cell_idx
                    elif elements:
                        # Element was removed — select the element at
                        # the same position (or last if at the end).
                        pos = getattr(self, '_saved_element_pos', 0) or 0
                        pos = min(pos, len(elements) - 1)
                        fallback_id = elements[pos][0]
                        self._select_element(
                            cell_idx,
                            fallback_id,
                            slot,
                            rule_id,
                        )
            self._saved_element = None
            self._saved_element_pos = None

    # ------------------------------------------------------------------
    # Per-element selection
    # ------------------------------------------------------------------

    def _clear_element_selection(self):
        """Clear the per-element selection and repaint the old cell."""
        if self._selected_element is not None:
            old_index = self._selected_index
            self._selected_element = None
            self._selected_index = QModelIndex()
            self._drag_start_pos = None
            if old_index.isValid():
                self.update(old_index)

    def _select_element(self, index, target_id, slot, rule_id):
        """Set the per-element selection and repaint affected cells."""
        old_index = self._selected_index
        self._selected_element = (rule_id, slot, target_id)
        self._selected_index = index
        if old_index.isValid() and old_index != index:
            self.update(old_index)
        self.update(index)

    def _select_element_at(self, index, vp_pos, model):
        """Select the element at *vp_pos* for any selectable column."""
        col = index.column()
        col_to_slot = model.col_to_slot
        element_cols = _model_element_cols(model)
        slot = col_to_slot.get(col)
        if not slot:
            return
        elem = self._element_at_pos(index, vp_pos)
        if elem is not None:
            tid, _n, _t, *_ = elem
            rd = model.get_row_data(index)
            if rd is not None:
                self._select_element(index, tid, slot, rd.rule_id)
        elif col not in element_cols:
            rd = model.get_row_data(index)
            if rd is not None:
                self._select_element(index, f'__{slot}__', slot, rd.rule_id)

    def _element_at_pos(self, index, viewport_pos):
        """Return the ``(target_id, name, type)`` element at *viewport_pos*, or ``None``."""
        elements = index.data(ELEMENTS_ROLE)
        if not elements:
            return None
        vrect = self.visualRect(index)
        if not vrect.isValid():
            return None
        delegate = self.itemDelegate()
        icon_sz = delegate._icon_size()
        fm = self.fontMetrics()
        line_h = max(icon_sz, fm.height())
        relative_y = viewport_pos.y() - vrect.top() - delegate._V_PAD
        if relative_y < 0:
            return None
        elem_index = int(relative_y // line_h)
        if 0 <= elem_index < len(elements):
            return elements[elem_index]
        return None

    # ------------------------------------------------------------------
    # Current-index tracking (position column highlight)
    # ------------------------------------------------------------------

    def currentChanged(self, current, previous):
        super().currentChanged(current, previous)
        model = self.model()
        if model is None:
            return
        pos_col = model.position_col
        # Repaint the position column cell for the old and new row so the
        # "row header" highlight follows the cursor (like Excel).
        for idx in (previous, current):
            if idx.isValid():
                pos_idx = model.index(idx.row(), pos_col, idx.parent())
                self.update(pos_idx)

    # ------------------------------------------------------------------
    # Mouse events for per-element selection and drag
    # ------------------------------------------------------------------

    def mousePressEvent(self, event):
        self.clear_highlight()
        if event.button() == Qt.MouseButton.LeftButton:
            pos = event.position().toPoint()
            index = self.indexAt(pos)
            model = self.model()
            selectable = _model_selectable_cols(model)
            element_cols = _model_element_cols(model)
            col_to_slot = model.col_to_slot if model is not None else {}
            if (
                index.isValid()
                and model is not None
                and not model.is_group(index)
                and index.column() in selectable
            ):
                col = index.column()
                elem = self._element_at_pos(index, pos)
                if elem is not None:
                    target_id, _name, _obj_type, *_ = elem
                    slot = col_to_slot.get(col)
                    row_data = model.get_row_data(index)
                    if row_data is not None and slot:
                        self._select_element(
                            index,
                            target_id,
                            slot,
                            row_data.rule_id,
                        )
                        # Only start drag for true element columns.
                        if col in element_cols:
                            self._drag_start_pos = pos
                    else:
                        self._clear_element_selection()
                elif col not in element_cols:
                    # Non-element columns (Action, Direction, Options)
                    # are always selectable even when the cell is empty.
                    slot = col_to_slot.get(col)
                    row_data = model.get_row_data(index)
                    if row_data is not None and slot:
                        self._select_element(
                            index,
                            f'__{slot}__',
                            slot,
                            row_data.rule_id,
                        )
                    else:
                        self._clear_element_selection()
                else:
                    self._clear_element_selection()
            else:
                self._clear_element_selection()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if (
            self._drag_start_pos is not None
            and self._selected_element is not None
            and event.buttons() & Qt.MouseButton.LeftButton
        ):
            dist = (event.position().toPoint() - self._drag_start_pos).manhattanLength()
            if dist >= QApplication.startDragDistance():
                self._start_element_drag()
                return
        super().mouseMoveEvent(event)

    def _start_element_drag(self):
        """Initiate a drag operation for the selected element."""
        rule_id, slot, target_id = self._selected_element
        model = self.model()
        if model is None:
            return
        row_data = model.get_row_data(self._selected_index)
        if row_data is None:
            return
        elements = getattr(row_data, slot, [])
        name = ''
        obj_type = ''
        for eid, ename, etype, *_ in elements:
            if eid == target_id:
                name = ename
                obj_type = etype
                break
        if not obj_type:
            return

        payload = json.dumps(
            {
                'id': str(target_id),
                'name': name,
                'source_rule_id': str(rule_id),
                'source_slot': slot,
                'type': obj_type,
            }
        ).encode()

        mime_data = QMimeData()
        mime_data.setData(FWF_MIME_TYPE, payload)

        drag = QDrag(self)
        drag.setMimeData(mime_data)

        icon_sz = QSettings().value('UI/IconSizeInRules', 25, type=int)
        icon = QIcon(f':/Icons/{obj_type}/icon')
        if not icon.isNull():
            drag.setPixmap(icon.pixmap(icon_sz, icon_sz))

        self._drag_start_pos = None
        drag.exec(Qt.DropAction.CopyAction | Qt.DropAction.MoveAction)

    # ------------------------------------------------------------------
    # Double-click to rename group
    # ------------------------------------------------------------------

    def _on_double_clicked(self, index):
        model = self.model()
        if model is None:
            return
        if model.is_group(index):
            old_name = model.group_name(index)
            new_name, ok = QInputDialog.getText(
                self,
                'Rename Group',
                'Group name:',
                text=old_name,
            )
            if ok and new_name and new_name != old_name:
                model.rename_group(index, new_name)
            return

        col = index.column()
        row_data = model.get_row_data(index)
        if row_data is None:
            return

        element_cols = _model_element_cols(model)
        col_to_slot = model.col_to_slot
        action_col = model.action_col
        comment_col = model.comment_col
        direction_col = model.direction_col
        metric_col = model.metric_col
        options_col = model.options_col

        if col in element_cols:
            slot = col_to_slot.get(col)
            if not slot:
                return
            elements = getattr(row_data, slot, [])
            if elements:
                target = None
                if self._selected_element is not None:
                    sel_rid, sel_slot, sel_tid = self._selected_element
                    if sel_rid == row_data.rule_id and sel_slot == slot:
                        for eid, ename, etype, *_ in elements:
                            if eid == sel_tid:
                                target = (eid, ename, etype)
                                break
                if target is None:
                    target = elements[0]
                if target[1] == 'Any':
                    self._show_any_message(col)
                else:
                    self._open_element_editor(str(target[0]), target[2])
                    self._reveal_in_tree(str(target[0]))
        elif col == action_col:
            self._open_action_editor(model, index)
        elif col == comment_col:
            self._open_comment_editor(model, index)
        elif col == direction_col:
            self._open_direction_editor(model, index)
        elif col == metric_col:
            self._open_metric_editor(model, index)
        elif col == options_col:
            self._open_rule_options_dialog(model, index)

    # ------------------------------------------------------------------
    # Context menu
    # ------------------------------------------------------------------

    def contextMenuEvent(self, event):
        """Show context menu, using viewport coordinates for correct hit-testing."""
        vp_pos = self.viewport().mapFromGlobal(event.globalPos())
        index = self.indexAt(vp_pos)
        global_pos = event.globalPos()
        model = self.model()
        if model is None:
            return

        menu = QMenu(self)

        if not index.isValid():
            # Clicked on empty area.
            menu.addAction(
                'Insert New Rule on Top',
                lambda: self._insert_and_scroll(model, at_top=True),
            )
            menu.addAction(
                'Insert New Rule at Bottom',
                lambda: self._insert_and_scroll(model, at_bottom=True),
            )
            menu.exec(global_pos)
            return

        # setFirstColumnSpanned on group headers can cause indexAt() to
        # return the group index even for clicks on child rule rows.
        # Detect this and resolve to the correct child index.
        if model.is_group(index):
            child_index = self._child_index_at(model, index, vp_pos)
            if child_index is not None:
                index = child_index
            else:
                build_group_header_menu(menu, self, model, index)
                menu.exec(global_pos)
                return

        col = index.column()
        element_cols = _model_element_cols(model)
        col_to_slot = model.col_to_slot
        selectable = _model_selectable_cols(model)
        action_col = model.action_col
        comment_col = model.comment_col
        direction_col = model.direction_col
        options_col = model.options_col

        # Element and comment columns get their own self-contained menus
        # without group/color actions.
        if col in element_cols:
            # Hit-test for per-element selection on right-click.
            elem = self._element_at_pos(index, vp_pos)
            if elem is not None:
                tid, _n, _t, *_ = elem
                rd = model.get_row_data(index)
                slot = col_to_slot.get(col)
                if rd is not None and slot:
                    self._select_element(index, tid, slot, rd.rule_id)
            build_element_menu(menu, self, model, index, col)
            menu.exec(global_pos)
            return

        if col == comment_col:
            build_comment_menu(menu, self, model, index)
            menu.exec(global_pos)
            return

        # Columns with self-contained menus (no group/color).
        # Select the element on right-click for visual feedback.
        if col in selectable and col not in element_cols:
            self._select_element_at(index, vp_pos, model)

        if col == action_col:
            build_action_menu(menu, self, model, index)
            menu.exec(global_pos)
            return

        if col == direction_col:
            build_direction_menu(menu, self, model, index)
            menu.exec(global_pos)
            return

        metric_col = model.metric_col
        if col == metric_col:
            build_metric_menu(menu, self, model, index)
            menu.exec(global_pos)
            return

        if col == options_col:
            build_options_menu(menu, self, model, index)
            menu.exec(global_pos)
            return

        # Remaining columns get group/color actions.
        row_data = model.get_row_data(index)
        in_group = row_data is not None and row_data.group

        if in_group:
            selected = self._selected_rule_indices()
            if not selected:
                selected = [index]
            # Collect all selected in-group rule IDs; show action if any
            # selected rule is outermost (matches fwbuilder behavior).
            group_rids = []
            any_outermost = False
            for sel in selected:
                rd = model.get_row_data(sel)
                if rd is not None and rd.group:
                    group_rids.append(rd.rule_id)
                    if model.is_outermost(sel):
                        any_outermost = True
            if any_outermost and group_rids:
                menu.addAction(
                    'Remove From Group',
                    lambda rids=group_rids: model.remove_from_group_by_ids(rids),
                )
                menu.addSeparator()
        else:
            add_new_group_action(menu, self, model, index)
            add_to_adjacent_group_actions(menu, self, model, index)

        add_color_submenu(menu, self, model, index)
        menu.addSeparator()

        build_row_menu(menu, self, model, index)

        menu.exec(global_pos)

    def _child_index_at(self, model, group_index, vp_pos):
        """Resolve a click below a group header to the correct child index.

        When ``setFirstColumnSpanned`` is active, ``indexAt()`` may return
        the group index for clicks on child rows.  This method checks the
        visual rects to find the actual child.

        Returns a child :class:`QModelIndex`, or *None* if the click is on
        the group header itself.
        """
        group_rect = self.visualRect(group_index)
        if group_rect.isValid() and group_rect.contains(vp_pos):
            return None  # Click is genuinely on the group header.

        child_count = model.rowCount(group_index)
        for row in range(child_count):
            child_idx = model.index(row, 0, group_index)
            child_rect = self.visualRect(child_idx)
            if (
                child_rect.isValid()
                and child_rect.top() <= vp_pos.y() <= child_rect.bottom()
            ):
                # Determine the correct column from the x coordinate.
                for col in range(model.columnCount()):
                    col_idx = model.index(row, col, group_index)
                    col_rect = self.visualRect(col_idx)
                    if (
                        col_rect.isValid()
                        and col_rect.left() <= vp_pos.x() <= col_rect.right()
                    ):
                        return col_idx
                return child_idx  # Fallback: column-0 child.
        return None

    def _rename_group_dialog(self, model, group_index):
        old_name = model.group_name(group_index)
        new_name, ok = QInputDialog.getText(
            self,
            'Rename Group',
            'Group name:',
            text=old_name,
        )
        if ok and new_name and new_name != old_name:
            model.rename_group(group_index, new_name)

    def _do_compile_rule(self, model, index):
        """Trigger single-rule compilation via the main window."""
        row_data = model.get_row_data(index)
        if row_data is None:
            return
        if self._bridge is not None:
            self._bridge.compile_single_rule(row_data.rule_id, model.rule_set_id)

    def _open_element_editor(self, obj_id, obj_type):
        """Open the object editor for the given element."""
        if self._bridge is not None:
            self._bridge.open_object_editor(obj_id, obj_type)

    def _show_any_message(self, col):
        """Show the 'Any' object description in the editor pane."""
        if self._bridge is not None:
            model = self.model()
            slot = model.col_to_slot.get(col, '') if model else ''
            self._bridge.show_any_editor(slot)

    def _reveal_in_tree(self, obj_id):
        """Select and reveal the object in the object tree."""
        if self._bridge is not None:
            self._bridge.reveal_in_tree(obj_id)

    def _show_where_used(self, obj_id, name, obj_type):
        """Show where-used results for the given object."""
        if self._bridge is not None:
            self._bridge.show_where_used(obj_id, name, obj_type)

    def _copy_element(self, target_id, name, obj_type):
        """Copy the element to the object clipboard."""
        self._clipboard_store.set_object(str(target_id), name, obj_type)

    def _cut_element(self, model, index, slot, target_id, name, obj_type):
        """Copy the element to clipboard and remove it from the cell."""
        self._copy_element(target_id, name, obj_type)
        model.remove_element(index, slot, target_id)

    def _paste_element(self, model, index, slot):
        """Paste the object clipboard into the cell."""
        entry = self._clipboard_store.object_entry
        if entry is None:
            return
        model.add_element(index, slot, uuid.UUID(entry['id']))

    def _change_action_and_edit(self, model, index, action):
        """Change the rule action and open the action editor with a fresh index."""
        row_data = model.get_row_data(index)
        if row_data is None:
            return
        rule_id = row_data.rule_id
        model.set_action(index, action)
        # set_action() triggers reload(), invalidating the old index.
        new_index = model.index_for_rule(rule_id)
        if new_index.isValid():
            self._open_action_editor(model, new_index)

    def _open_action_editor(self, model, index):
        """Open the Action parameters panel in the editor pane."""
        if self._bridge is not None:
            self._bridge.open_action_editor(model, index)

    def _open_comment_editor(self, model, index):
        """Open the Comment editor panel in the editor pane."""
        if self._bridge is not None:
            self._bridge.open_comment_editor(model, index)

    def _open_direction_editor(self, model, index):
        """Open the (blank) Direction pane in the editor pane."""
        if self._bridge is not None:
            self._bridge.open_direction_editor(model, index)

    def _open_metric_editor(self, model, index):
        """Open the Metric editor panel in the editor pane."""
        if self._bridge is not None:
            self._bridge.open_metric_editor(model, index)

    def _open_rule_options_dialog(self, model, index):
        """Open the Rule Options panel in the editor pane."""
        row_data = model.get_row_data(index)
        if row_data is None:
            return
        if self._bridge is not None:
            self._bridge.open_rule_options(model, index)

    @staticmethod
    def _set_disabled_on_selection(model, indices, *, disabled):
        """Set the disabled state for all rules in *indices*."""
        for idx in indices:
            model.set_disabled(idx, disabled)

    @staticmethod
    def _set_label_on_selection(model, indices, label_key):
        """Apply *label_key* to all rules in *indices*."""
        for idx in indices:
            model.set_label(idx, label_key)

    # ------------------------------------------------------------------
    # Public clipboard actions (called from main window edit menu)
    # ------------------------------------------------------------------

    def copy_selection(self):
        """Copy selected rules to the clipboard."""
        model = self.model()
        if model is None:
            return
        indices = self._selected_rule_indices()
        if indices:
            model.copy_rules(indices)

    def cut_selection(self):
        """Cut selected rules to the clipboard."""
        model = self.model()
        if model is None:
            return
        indices = self._selected_rule_indices()
        if indices:
            model.cut_rules(indices)

    def copy_object(self):
        """Copy element or rules -- context-aware like fwbuilder.

        If a single element is selected in an element column, copy that
        element to the object clipboard.  Otherwise copy whole rules.
        """
        if self._selected_element is not None:
            _rule_id, _slot, target_id = self._selected_element
            idx = self._selected_index
            model = self.model()
            element_cols = _model_element_cols(model)
            if idx.isValid() and idx.column() in element_cols:
                elements = idx.data(ELEMENTS_ROLE) or []
                for eid, ename, etype, *_ in elements:
                    if eid == target_id:
                        self._copy_element(target_id, ename, etype)
                        return
        self.copy_selection()

    def cut_object(self):
        """Cut element or rules -- context-aware like fwbuilder.

        If a single element is selected in an element column, cut that
        element.  Otherwise cut whole rules.
        """
        if self._selected_element is not None:
            _rule_id, slot, target_id = self._selected_element
            idx = self._selected_index
            model = self.model()
            element_cols = _model_element_cols(model)
            if idx.isValid() and idx.column() in element_cols and model is not None:
                elements = idx.data(ELEMENTS_ROLE) or []
                for eid, ename, etype, *_ in elements:
                    if eid == target_id:
                        self._cut_element(model, idx, slot, target_id, ename, etype)
                        return
        self.cut_selection()

    def delete_selection(self):
        """Delete element or rules -- context-aware like fwbuilder.

        If a single element is selected in an element column, remove
        that element from the cell.  If the current cell is in an
        element column but no element is individually selected, do
        nothing (don't accidentally delete the whole rule).  Only
        delete whole rules when the current cell is outside element
        columns (e.g. Position or Comment).
        """
        model = self.model()
        element_cols = _model_element_cols(model)
        if self._selected_element is not None:
            _rule_id, slot, target_id = self._selected_element
            idx = self._selected_index
            if idx.isValid() and idx.column() in element_cols and model is not None:
                model.remove_element(idx, slot, target_id)
                return
        # Guard: if the current cell is in an element column, don't
        # fall through to deleting the whole rule -- the user intended
        # to remove an element, not the row.
        cur = self.currentIndex()
        if cur.isValid() and cur.column() in element_cols:
            return
        if model is None:
            return
        indices = self._selected_rule_indices()
        if indices:
            model.delete_rules(indices)

    def paste_below(self):
        """Paste rules from clipboard below the current rule."""
        model = self.model()
        if model is None:
            return
        idx = self.currentIndex()
        if idx.isValid() and not model.is_group(idx):
            self._paste_and_scroll(model, idx)

    def paste_object(self):
        """Paste clipboard content -- object into cell or rules below.

        Mimics fwbuilder's ``RuleSetView::pasteObject()``: if the
        object clipboard holds a regular object and the current cell
        can accept it, paste it there; otherwise fall back to rule
        paste.
        """
        model = self.model()
        if model is None:
            return
        idx = self.currentIndex()
        if not idx.isValid() or model.is_group(idx):
            return

        # If the object clipboard has a compatible object for the
        # current cell's slot, paste the object into that cell.
        col = idx.column()
        element_cols = _model_element_cols(model)
        col_to_slot = model.col_to_slot
        if col in element_cols:
            slot = col_to_slot.get(col)
            if slot:
                valid_types = VALID_TYPES_BY_SLOT.get(slot, frozenset())
                entry = self._clipboard_store.object_entry
                if entry is not None and entry.get('type', '') in valid_types:
                    self._paste_element(model, index=idx, slot=slot)
                    return

        # Fall back to rule paste.
        self._paste_and_scroll(model, idx)

    # ------------------------------------------------------------------
    # Selection helpers
    # ------------------------------------------------------------------

    def _selected_rule_indices(self):
        """Return unique QModelIndex list for selected rule rows."""
        seen = set()
        result = []
        for idx in self.selectedIndexes():
            model = self.model()
            if model is None:
                continue
            # Normalize to column 0.
            idx0 = model.index(idx.row(), 0, idx.parent())
            key = (idx0.row(), idx0.internalPointer())
            if key not in seen and not model.is_group(idx0):
                seen.add(key)
                result.append(idx0)
        return result

    def _insert_and_scroll(self, model, index=None, **kwargs):
        """Insert a rule and scroll to it."""
        rule_id = model.insert_rule(index, **kwargs)
        if rule_id is not None:
            new_idx = model.index_for_rule(rule_id)
            if new_idx.isValid():
                self.scrollTo(new_idx)
                self.setCurrentIndex(new_idx)

    def _paste_and_scroll(self, model, index, *, before=False):
        """Paste rules from clipboard and scroll to the first pasted rule."""
        new_ids = model.paste_rules(index, before=before)
        if new_ids:
            first_idx = model.index_for_rule(new_ids[0])
            if first_idx.isValid():
                self.scrollTo(first_idx)
                self.setCurrentIndex(first_idx)

    def _move_and_select(self, rule_id):
        """Re-select the moved rule after a move operation."""
        if rule_id is None:
            return
        model = self.model()
        if model is None:
            return
        new_idx = model.index_for_rule(rule_id)
        if new_idx.isValid():
            self.setCurrentIndex(new_idx)

    # ------------------------------------------------------------------
    # Keyboard shortcuts
    # ------------------------------------------------------------------

    def keyPressEvent(self, event):
        model = self.model()
        if model is None:
            super().keyPressEvent(event)
            return

        key = event.key()
        modifiers = event.modifiers()

        if modifiers == Qt.KeyboardModifier.ControlModifier:
            if key == Qt.Key.Key_C:
                self.copy_object()
                return
            if key == Qt.Key.Key_V:
                self.paste_object()
                return
            if key == Qt.Key.Key_X:
                self.cut_object()
                return
            if key == Qt.Key.Key_PageUp:
                idx = self.currentIndex()
                if idx.isValid() and not model.is_group(idx):
                    self._move_and_select(model.move_rule_up(idx))
                return
            if key == Qt.Key.Key_PageDown:
                idx = self.currentIndex()
                if idx.isValid() and not model.is_group(idx):
                    self._move_and_select(model.move_rule_down(idx))
                return

        if modifiers == Qt.KeyboardModifier.NoModifier:
            if key == Qt.Key.Key_Delete:
                self.delete_selection()
                return
            if key == Qt.Key.Key_Insert:
                idx = self.currentIndex()
                if idx.isValid() and not model.is_group(idx):
                    self._insert_and_scroll(model, index=idx)
                else:
                    self._insert_and_scroll(model, at_bottom=True)
                return
            if key == Qt.Key.Key_X:
                idx = self.currentIndex()
                if idx.isValid() and not model.is_group(idx):
                    self._do_compile_rule(model, idx)
                return

        super().keyPressEvent(event)

    # ------------------------------------------------------------------
    # Drop support
    # ------------------------------------------------------------------

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat(FWF_MIME_TYPE):
            event.acceptProposedAction()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if not event.mimeData().hasFormat(FWF_MIME_TYPE):
            event.ignore()
            return
        index = self.indexAt(event.position().toPoint())
        model = self.model()
        element_cols = _model_element_cols(model)
        if (
            index.isValid()
            and index.column() in element_cols
            and model is not None
            and not model.is_group(index)
        ):
            if event.keyboardModifiers() & Qt.KeyboardModifier.ControlModifier:
                event.setDropAction(Qt.DropAction.CopyAction)
            else:
                event.setDropAction(Qt.DropAction.MoveAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        mime = event.mimeData()
        if not mime.hasFormat(FWF_MIME_TYPE):
            event.ignore()
            return

        index = self.indexAt(event.position().toPoint())
        model = self.model()
        element_cols = _model_element_cols(model)
        col_to_slot = model.col_to_slot if model is not None else {}
        if (
            not index.isValid()
            or index.column() not in element_cols
            or model is None
            or model.is_group(index)
        ):
            event.ignore()
            return

        try:
            payload = json.loads(bytes(mime.data(FWF_MIME_TYPE)).decode())
        except (json.JSONDecodeError, UnicodeDecodeError):
            event.ignore()
            return

        # Normalise payload: old format (single dict) → list.
        if isinstance(payload, dict):
            items = [payload]
        elif isinstance(payload, list):
            items = payload
        else:
            event.ignore()
            return

        if not items:
            event.ignore()
            return

        slot = col_to_slot.get(index.column())
        if not slot:
            event.ignore()
            return
        valid_types = VALID_TYPES_BY_SLOT.get(slot, frozenset())

        # Cell-to-cell drag (always single item).
        first = items[0]
        source_rule_id = first.get('source_rule_id')
        source_slot = first.get('source_slot')

        if source_rule_id and source_slot:
            obj_id = first.get('id')
            obj_type = first.get('type', '')
            if not obj_id or obj_type not in valid_types:
                event.ignore()
                return
            row_data = model.get_row_data(index)
            if row_data is None:
                event.ignore()
                return
            if str(row_data.rule_id) == source_rule_id and source_slot == slot:
                event.ignore()
                return
            if event.keyboardModifiers() & Qt.KeyboardModifier.ControlModifier:
                model.add_element(index, slot, uuid.UUID(obj_id))
            else:
                model.move_element(
                    uuid.UUID(source_rule_id),
                    source_slot,
                    index,
                    slot,
                    uuid.UUID(obj_id),
                )
        else:
            # Tree drop — add all valid items.
            for entry in items:
                obj_id = entry.get('id')
                obj_type = entry.get('type', '')
                if obj_id and obj_type in valid_types:
                    model.add_element(index, slot, uuid.UUID(obj_id))

        event.acceptProposedAction()


class RuleSetPanel(QWidget):
    """Container with title label, 'Insert Rule' button, and a :class:`PolicyView`.

    Matches fwbuilder's ``projectpanel_q.ui`` toolbar layout: [+] button,
    spacer, centred title label (14pt), spacer.
    """

    def __init__(self, *, clipboard_store=None, bridge=None, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        toolbar = QHBoxLayout()
        toolbar.setContentsMargins(2, 2, 2, 2)
        add_btn = QToolButton()
        add_btn.setIcon(QIcon(':/Icons/Add'))
        add_btn.setIconSize(QSize(25, 25))
        add_btn.setToolTip('Insert rule')
        add_btn.clicked.connect(self._insert_rule)
        toolbar.addWidget(add_btn)

        self._title_label = QLabel()
        font = QFont()
        font.setPointSize(14)
        self._title_label.setFont(font)
        self._title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        toolbar.addWidget(self._title_label, 1)

        layout.addLayout(toolbar)

        self.policy_view = PolicyView(clipboard_store=clipboard_store, bridge=bridge)
        layout.addWidget(self.policy_view)

    def set_title(self, title):
        """Set the title label text (e.g. 'fw-name / Policy')."""
        self._title_label.setText(title)

    def _insert_rule(self):
        model = self.policy_view.model()
        if model is None:
            return
        idx = self.policy_view.currentIndex()
        if idx.isValid() and not model.is_group(idx):
            self.policy_view._insert_and_scroll(model, index=idx, before=True)
        else:
            self.policy_view._insert_and_scroll(model, at_bottom=True)



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Jinja2 template loader and renderer.

Follows the same resource resolution pattern as ``Configlet``:
check ``~/firewallfabrik/templates/<platform>/`` for user overrides
first, fall back to the package's ``resources/templates/<platform>/``
directory.
"""

from __future__ import annotations

import importlib.resources
from pathlib import Path

import jinja2


def _get_package_resources_dir() -> Path:
    """Return the path to the package's resources directory."""
    ref = importlib.resources.files('firewallfabrik') / 'resources'
    return Path(str(ref))


class Jinja2Template:
    """Load and render a Jinja2 template by platform and name."""

    def __init__(self, platform: str, template_name: str) -> None:
        search_paths: list[str] = []

        # User override directory (checked first)
        user_dir = Path.home() / 'firewallfabrik' / 'templates' / platform
        if user_dir.is_dir():
            search_paths.append(str(user_dir))

        # Package resources directory (fallback)
        pkg_dir = _get_package_resources_dir() / 'templates' / platform
        search_paths.append(str(pkg_dir))

        loader = jinja2.FileSystemLoader(search_paths)
        self._env = jinja2.Environment(
            loader=loader,
            undefined=jinja2.StrictUndefined,
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
        )
        self._template = self._env.get_template(template_name)

    def render(self, context: dict) -> str:
        """Render the template with the given context variables."""
        return self._template.render(context)



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Custom QUiLoader that populates an existing widget instead of creating a new one."""

from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QDockWidget, QTextBrowser, QWidget

# Maps custom widget class names from the .ui file to their Qt base classes.
# As Python implementations are created, replace QWidget with the real class.
CUSTOM_WIDGET_MAP = {
    'ActionsDialog': QWidget,
    'AddressRangeDialog': QWidget,
    'AddressTableDialog': QWidget,
    'AttachedNetworksDialog': QWidget,
    'BlankDialog': QWidget,
    'ClusterDialog': QWidget,
    'ClusterGroupDialog': QWidget,
    'CommentEditorPanel': QWidget,
    'CommentKeywords': QWidget,  # replaced by CommentTags at module bottom
    'CompilerOutputPanel': QTextBrowser,
    'CustomServiceDialog': QWidget,
    'DNSNameDialog': QWidget,
    'DynamicGroupDialog': QWidget,
    'FirewallDialog': QWidget,
    'GroupObjectDialog': QWidget,
    'HostDialog': QWidget,
    'ICMPServiceDialog': QWidget,
    'InterfaceDialog': QWidget,
    'IPServiceDialog': QWidget,
    'IPv4Dialog': QWidget,
    'IPv6Dialog': QWidget,
    'LibraryDialog': QWidget,
    'MetricEditorPanel': QWidget,
    'NATRuleOptionsDialog': QWidget,
    'NetworkDialog': QWidget,
    'NetworkDialogIPv6': QWidget,
    'ObjectEditorDockWidget': QDockWidget,
    'PhysicalAddressDialog': QWidget,
    'RoutingRuleOptionsDialog': QWidget,
    'RuleOptionsDialog': QWidget,
    'RuleSetDialog': QWidget,  # replaced by RuleSetDialog at module bottom
    'TagServiceDialog': QWidget,
    'TCPServiceDialog': QWidget,
    'TimeDialog': QWidget,
    'UDPServiceDialog': QWidget,
    'UserDialog': QWidget,
}


class FWFUiLoader(QUiLoader):
    """Custom UI loader that populates an existing widget instance.

    This mimics the C++ ``Ui::setupUi(this)`` pattern.  When
    ``QUiLoader`` creates the top-level widget (parent is ``None``),
    we return the *base_instance* so that all child widgets, menus,
    toolbars, and dock widgets are added directly to it.

    Unknown custom widget classes listed in ``CUSTOM_WIDGET_MAP`` are
    replaced with their Qt base class so .ui files can be loaded before
    all Python widget classes exist.
    """

    def __init__(self, base_instance):
        super().__init__(base_instance)
        self._base_instance = base_instance

    def createWidget(self, class_name, parent=None, name=''):
        if parent is None and self._base_instance is not None:
            return self._base_instance
        if class_name in CUSTOM_WIDGET_MAP:
            widget = CUSTOM_WIDGET_MAP[class_name](parent)
            widget.setObjectName(name)
            return widget
        return super().createWidget(class_name, parent, name)


# Late imports: dialog modules import FWFUiLoader from this module,
# so we update the map after both modules are fully defined.
def _register_actions_dialog():
    from firewallfabrik.gui.actions_dialog import ActionsPanel

    CUSTOM_WIDGET_MAP['ActionsDialog'] = ActionsPanel


def _register_address_dialogs():
    from firewallfabrik.gui.address_dialogs import (
        AddressRangeDialog,
        IPv4Dialog,
        IPv6Dialog,
        NetworkDialog,
        NetworkDialogIPv6,
        PhysAddressDialog,
    )

    CUSTOM_WIDGET_MAP['IPv4Dialog'] = IPv4Dialog
    CUSTOM_WIDGET_MAP['IPv6Dialog'] = IPv6Dialog
    CUSTOM_WIDGET_MAP['NetworkDialog'] = NetworkDialog
    CUSTOM_WIDGET_MAP['NetworkDialogIPv6'] = NetworkDialogIPv6
    CUSTOM_WIDGET_MAP['AddressRangeDialog'] = AddressRangeDialog
    CUSTOM_WIDGET_MAP['PhysicalAddressDialog'] = PhysAddressDialog


def _register_device_dialogs():
    from firewallfabrik.gui.device_dialogs import (
        FirewallDialog,
        HostDialog,
        InterfaceDialog,
    )

    CUSTOM_WIDGET_MAP['HostDialog'] = HostDialog
    CUSTOM_WIDGET_MAP['FirewallDialog'] = FirewallDialog
    CUSTOM_WIDGET_MAP['InterfaceDialog'] = InterfaceDialog


def _register_service_dialogs():
    from firewallfabrik.gui.service_dialogs import (
        ICMPServiceDialog,
        IPServiceDialog,
        TCPServiceDialog,
        UDPServiceDialog,
    )

    CUSTOM_WIDGET_MAP['TCPServiceDialog'] = TCPServiceDialog
    CUSTOM_WIDGET_MAP['UDPServiceDialog'] = UDPServiceDialog
    CUSTOM_WIDGET_MAP['ICMPServiceDialog'] = ICMPServiceDialog
    CUSTOM_WIDGET_MAP['IPServiceDialog'] = IPServiceDialog


def _register_group_dialog():
    from firewallfabrik.gui.group_dialog import GroupObjectDialog

    CUSTOM_WIDGET_MAP['GroupObjectDialog'] = GroupObjectDialog


def _register_group_type_dialogs():
    from firewallfabrik.gui.address_table_dialog import AddressTableDialog
    from firewallfabrik.gui.dns_name_dialog import DNSNameDialog
    from firewallfabrik.gui.dynamic_group_dialog import DynamicGroupDialog

    CUSTOM_WIDGET_MAP['AddressTableDialog'] = AddressTableDialog
    CUSTOM_WIDGET_MAP['DNSNameDialog'] = DNSNameDialog
    CUSTOM_WIDGET_MAP['DynamicGroupDialog'] = DynamicGroupDialog


def _register_library_dialog():
    from firewallfabrik.gui.library_dialog import LibraryDialog

    CUSTOM_WIDGET_MAP['LibraryDialog'] = LibraryDialog


def _register_time_dialog():
    from firewallfabrik.gui.time_dialog import TimeDialog

    CUSTOM_WIDGET_MAP['TimeDialog'] = TimeDialog


def _register_comment_tags():
    from firewallfabrik.gui.comment_tags import CommentTags

    CUSTOM_WIDGET_MAP['CommentKeywords'] = CommentTags


def _register_comment_editor_panel():
    from firewallfabrik.gui.comment_editor_panel import CommentEditorPanel

    CUSTOM_WIDGET_MAP['CommentEditorPanel'] = CommentEditorPanel


def _register_drop_area():
    from firewallfabrik.gui.drop_area import FWObjectDropArea

    CUSTOM_WIDGET_MAP['FWObjectDropArea'] = FWObjectDropArea


def _register_metric_editor_panel():
    from firewallfabrik.gui.metric_editor_panel import MetricEditorPanel

    CUSTOM_WIDGET_MAP['MetricEditorPanel'] = MetricEditorPanel


def _register_nat_rule_options_panel():
    from firewallfabrik.gui.nat_rule_options_dialog import NATRuleOptionsPanel

    CUSTOM_WIDGET_MAP['NATRuleOptionsDialog'] = NATRuleOptionsPanel


def _register_routing_rule_options_panel():
    from firewallfabrik.gui.routing_rule_options_dialog import RoutingRuleOptionsPanel

    CUSTOM_WIDGET_MAP['RoutingRuleOptionsDialog'] = RoutingRuleOptionsPanel


def _register_rule_options_panel():
    from firewallfabrik.gui.rule_options_dialog import RuleOptionsPanel

    CUSTOM_WIDGET_MAP['RuleOptionsDialog'] = RuleOptionsPanel


def _register_service_type_dialogs():
    from firewallfabrik.gui.custom_service_dialog import CustomServiceDialog
    from firewallfabrik.gui.tag_service_dialog import TagServiceDialog
    from firewallfabrik.gui.user_service_dialog import UserServiceDialog

    CUSTOM_WIDGET_MAP['CustomServiceDialog'] = CustomServiceDialog
    CUSTOM_WIDGET_MAP['TagServiceDialog'] = TagServiceDialog
    CUSTOM_WIDGET_MAP['UserDialog'] = UserServiceDialog


def _register_ruleset_dialog():
    from firewallfabrik.gui.ruleset_dialog import RuleSetDialog

    CUSTOM_WIDGET_MAP['RuleSetDialog'] = RuleSetDialog


_register_actions_dialog()
_register_address_dialogs()
_register_comment_editor_panel()
_register_comment_tags()
_register_device_dialogs()
_register_drop_area()
_register_group_dialog()
_register_group_type_dialogs()
_register_library_dialog()
_register_metric_editor_panel()
_register_nat_rule_options_panel()
_register_routing_rule_options_panel()
_register_rule_options_panel()
_register_ruleset_dialog()
_register_service_dialogs()
_register_service_type_dialogs()
_register_time_dialog()



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Action parameters editor panel for iptables policy rules."""

from pathlib import Path

from PySide6.QtWidgets import QWidget

from firewallfabrik.core.objects import PolicyAction
from firewallfabrik.gui.ui_loader import FWFUiLoader

# Reject type combo box items: (display_text, stored_value).
# The stored value is saved in options['action_on_reject'].
_REJECT_ITEMS = (
    ('', ''),
    ('ICMP admin prohibited', 'ICMP admin prohibited'),
    ('ICMP host prohibited', 'ICMP host prohibited'),
    ('ICMP host unreachable', 'ICMP host unreachable'),
    ('ICMP net prohibited', 'ICMP net prohibited'),
    ('ICMP net unreachable', 'ICMP net unreachable'),
    ('ICMP port unreachable', 'ICMP port unreachable'),
    ('ICMP protocol unreachable', 'ICMP protocol unreachable'),
    ('TCP RST', 'TCP RST'),
)

# Map action enum → stacked-widget page name.
_ACTION_PAGE = {
    PolicyAction.Accept: 'NonePage',
    PolicyAction.Accounting: 'AccountingStrPage',
    PolicyAction.Branch: 'BranchChainPage',
    PolicyAction.Continue: 'NonePage',
    PolicyAction.Custom: 'CustomStrPage',
    PolicyAction.Deny: 'NonePage',
    PolicyAction.Pipe: 'NonePage',
    PolicyAction.Reject: 'RejectPage',
}


class ActionsPanel(QWidget):
    """Editor-pane widget for action parameters (non-modal)."""

    def __init__(self, parent=None):
        super().__init__(parent)
        ui_path = Path(__file__).resolve().parent / 'ui' / 'actionsdialog_q.ui'
        loader = FWFUiLoader(self)
        loader.load(str(ui_path))

        self._model = None
        self._index = None
        self._rule_id = None
        self._loading = False
        self._signals_connected = False

        # Populate reject combo box.
        if hasattr(self, 'rejectvalue'):
            self.rejectvalue.clear()
            for display, _stored in _REJECT_ITEMS:
                self.rejectvalue.addItem(display)

    def load_rule(self, model, index):
        """Populate the panel from the rule at *index*."""
        self._disconnect_signals()
        self._model = model
        self._index = index
        row_data = model.get_row_data(index)
        self._rule_id = row_data.rule_id if row_data is not None else None
        self._load_options()
        self._connect_signals()

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _load_options(self):
        """Read options from the database and populate all widgets."""
        self._loading = True
        try:
            opts = self._read_rule_options()
            row_data = self._get_row_data()
            action_int = row_data.action_int if row_data is not None else 0

            # Switch to the correct page.
            try:
                action = PolicyAction(action_int)
            except (TypeError, ValueError):
                action = PolicyAction.Accept
            page_name = _ACTION_PAGE.get(action, 'NonePage')
            page = getattr(self, page_name, None)
            if page is not None and hasattr(self, 'widgetStack'):
                self.widgetStack.setCurrentWidget(page)

            # Reject page.
            if hasattr(self, 'rejectvalue'):
                val = str(opts.get('action_on_reject', ''))
                idx = self.rejectvalue.findText(val)
                self.rejectvalue.setCurrentIndex(idx if idx >= 0 else 0)

            # Accounting page.
            if hasattr(self, 'accountingvalue_str'):
                self.accountingvalue_str.setText(
                    str(opts.get('rule_name_accounting', '')),
                )

            # Custom page.
            if hasattr(self, 'custom_str'):
                self.custom_str.setText(str(opts.get('custom_str', '')))

            # Branch page.
            if hasattr(self, 'ipt_branch_in_mangle'):
                self.ipt_branch_in_mangle.setChecked(
                    _to_bool(opts.get('ipt_branch_in_mangle')),
                )
        finally:
            self._loading = False

    def _save_options(self):
        """Collect values from all widgets and persist via the model."""
        if self._model is None or self._index is None:
            return
        opts = self._read_rule_options()

        # Reject.
        if hasattr(self, 'rejectvalue'):
            opts['action_on_reject'] = self.rejectvalue.currentText()

        # Accounting.
        if hasattr(self, 'accountingvalue_str'):
            opts['rule_name_accounting'] = self.accountingvalue_str.text()

        # Custom.
        if hasattr(self, 'custom_str'):
            opts['custom_str'] = self.custom_str.text()

        # Branch.
        if hasattr(self, 'ipt_branch_in_mangle'):
            opts['ipt_branch_in_mangle'] = self.ipt_branch_in_mangle.isChecked()

        # Clean out empty/zero/false values to keep storage lean.
        cleaned = {}
        for k, v in opts.items():
            if v is None or v == '' or v == 0 or v is False:
                continue
            cleaned[k] = v

        self._model.set_options(self._index, cleaned)
        # set_options() calls reload(), invalidating all QModelIndex objects.
        # Re-resolve so subsequent saves use a valid index.
        if self._rule_id is not None:
            self._index = self._model.index_for_rule(self._rule_id)

    def _on_widget_changed(self):
        """Auto-save whenever any widget value changes."""
        if self._loading:
            return
        self._save_options()

    def _get_row_data(self):
        """Return the row data for the current index."""
        if self._model is None or self._index is None:
            return None
        return self._model.get_row_data(self._index)

    def _read_rule_options(self):
        """Read the full options dict from the database rule."""
        if self._model is None or self._index is None:
            return {}
        row_data = self._get_row_data()
        if row_data is None:
            return {}
        from firewallfabrik.core.objects import PolicyRule

        with self._model._db_manager.session() as session:
            rule = session.get(PolicyRule, row_data.rule_id)
            if rule is not None:
                return dict(rule.options or {})
        return {}

    # ------------------------------------------------------------------
    # Signal management
    # ------------------------------------------------------------------

    def _connect_signals(self):
        """Connect change signals on all widgets to auto-save."""
        if self._signals_connected:
            return
        if hasattr(self, 'rejectvalue'):
            self.rejectvalue.currentIndexChanged.connect(self._on_widget_changed)
        if hasattr(self, 'accountingvalue_str'):
            self.accountingvalue_str.editingFinished.connect(self._on_widget_changed)
        if hasattr(self, 'custom_str'):
            self.custom_str.editingFinished.connect(self._on_widget_changed)
        if hasattr(self, 'ipt_branch_in_mangle'):
            self.ipt_branch_in_mangle.toggled.connect(self._on_widget_changed)
        self._signals_connected = True

    def _disconnect_signals(self):
        """Disconnect all change signals to avoid stale callbacks."""
        if not self._signals_connected:
            return
        if hasattr(self, 'rejectvalue'):
            self.rejectvalue.currentIndexChanged.disconnect(self._on_widget_changed)
        if hasattr(self, 'accountingvalue_str'):
            self.accountingvalue_str.editingFinished.disconnect(
                self._on_widget_changed,
            )
        if hasattr(self, 'custom_str'):
            self.custom_str.editingFinished.disconnect(self._on_widget_changed)
        if hasattr(self, 'ipt_branch_in_mangle'):
            self.ipt_branch_in_mangle.toggled.disconnect(self._on_widget_changed)
        self._signals_connected = False


def _to_bool(val):
    """Convert a value to bool, handling string representations."""
    if isinstance(val, str):
        return val.lower() in ('true', '1')
    return bool(val)



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Editor panel dialogs for address objects."""

import ipaddress
import logging
import socket

from PySide6.QtCore import Slot
from PySide6.QtGui import QCursor, Qt
from PySide6.QtWidgets import QApplication, QMessageBox

from firewallfabrik.gui.base_object_dialog import BaseObjectDialog

logger = logging.getLogger(__name__)


def _resolve_name(hostname, af_type):
    """Resolve *hostname* to an IP address string using *af_type*.

    Returns the first matching address or an empty string on failure.
    """
    try:
        results = socket.getaddrinfo(hostname, None, af_type, socket.SOCK_STREAM)
        if results:
            return results[0][4][0]
    except (socket.gaierror, OSError):
        pass
    return ''


def _validate_ipv4(address_text):
    """Validate an IPv4 address string.  Return the normalised address or *None*."""
    try:
        return str(ipaddress.IPv4Address(address_text.strip()))
    except ValueError:
        return None


def _validate_ipv6(address_text):
    """Validate an IPv6 address string.  Return the normalised address or *None*."""
    try:
        return str(ipaddress.IPv6Address(address_text.strip()))
    except ValueError:
        return None


def _validate_ipv4_netmask(netmask_text):
    """Validate an IPv4 netmask (dotted-decimal or CIDR prefix length).

    Returns the normalised netmask string or *None*.
    """
    text = netmask_text.strip()
    if not text:
        return None
    # CIDR prefix length (e.g. "24").
    try:
        prefix = int(text)
        if 0 <= prefix <= 32:
            net = ipaddress.IPv4Network(f'0.0.0.0/{prefix}', strict=False)
            return str(net.netmask)
        return None
    except ValueError:
        pass
    # Dotted-decimal (e.g. "255.255.255.0").
    try:
        addr = ipaddress.IPv4Address(text)
        # Verify it's a valid contiguous netmask by round-tripping
        # through IPv4Network.
        ipaddress.IPv4Network(f'0.0.0.0/{addr}', strict=False)
        return str(addr)
    except ValueError:
        return None


class _BaseAddressDialog(BaseObjectDialog):
    """Base for IPv4/IPv6/Network/NetworkIPv6 dialogs (name + address + netmask)."""

    def _populate(self):
        inet = self._obj.inet_addr_mask or {}
        self.obj_name.setText(self._obj.name or '')
        self.address.setText(inet.get('address', ''))
        self.netmask.setText(inet.get('netmask', ''))

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        inet = dict(self._obj.inet_addr_mask or {})
        inet['address'] = self.address.text()
        inet['netmask'] = self.netmask.text()
        self._obj.inet_addr_mask = inet


class IPv4Dialog(_BaseAddressDialog):
    def __init__(self, parent=None):
        super().__init__('ipv4dialog_q.ui', parent)

    def _apply_changes(self):
        addr = self.address.text().strip()
        if addr and _validate_ipv4(addr) is None:
            QMessageBox.warning(
                self,
                self.tr('Invalid Address'),
                self.tr("Illegal IP address '%1'").replace('%1', addr),
            )
            return
        nm = self.netmask.text().strip()
        if nm and _validate_ipv4_netmask(nm) is None:
            QMessageBox.warning(
                self,
                self.tr('Invalid Netmask'),
                self.tr("Illegal netmask '%1'").replace('%1', nm),
            )
            return
        super()._apply_changes()

    @Slot()
    def addressEntered(self):
        """Parse CIDR notation (e.g. '192.168.1.1/24') into address + netmask."""
        text = self.address.text().strip()
        if '/' not in text:
            return
        try:
            net = ipaddress.IPv4Network(text, strict=False)
            self.address.setText(str(ipaddress.IPv4Address(text.split('/')[0])))
            self.netmask.setText(str(net.netmask))
        except ValueError:
            pass

    def _set_read_only(self, read_only):
        super()._set_read_only(read_only)
        self.dnsLookup.setEnabled(not read_only)

    @Slot()
    def DNSlookup(self):
        name = self.obj_name.text().strip()
        if not name:
            return
        QApplication.setOverrideCursor(QCursor(Qt.CursorShape.WaitCursor))
        addr = _resolve_name(name, socket.AF_INET)
        QApplication.restoreOverrideCursor()
        if addr:
            self.address.setText(addr)
            self._on_changed()
            return
        QMessageBox.warning(
            self,
            self.tr('DNS Lookup Failed'),
            self.tr("DNS lookup failed for '%1'.").replace('%1', name),
        )


class IPv6Dialog(_BaseAddressDialog):
    def __init__(self, parent=None):
        super().__init__('ipv6dialog_q.ui', parent)

    def _apply_changes(self):
        addr = self.address.text().strip()
        if addr and _validate_ipv6(addr) is None:
            QMessageBox.warning(
                self,
                self.tr('Invalid Address'),
                self.tr("Illegal IPv6 address '%1'").replace('%1', addr),
            )
            return
        nm = self.netmask.text().strip()
        if nm:
            try:
                prefix = int(nm)
                if prefix < 1 or prefix > 127:
                    raise ValueError
            except ValueError:
                QMessageBox.warning(
                    self,
                    self.tr('Invalid Netmask'),
                    self.tr("Illegal netmask '%1'").replace('%1', nm),
                )
                return
        super()._apply_changes()

    @Slot()
    def addressEntered(self):
        """Parse CIDR notation (e.g. '2001:db8::1/64') into address + prefix length."""
        text = self.address.text().strip()
        if '/' not in text:
            return
        try:
            net = ipaddress.IPv6Network(text, strict=False)
            self.address.setText(str(ipaddress.IPv6Address(text.split('/')[0])))
            self.netmask.setText(str(net.prefixlen))
        except ValueError:
            pass

    def _set_read_only(self, read_only):
        super()._set_read_only(read_only)
        self.dnsLookup.setEnabled(not read_only)

    @Slot()
    def changed(self):
        self._on_changed()

    @Slot()
    def DNSlookup(self):
        name = self.obj_name.text().strip()
        if not name:
            return
        QApplication.setOverrideCursor(QCursor(Qt.CursorShape.WaitCursor))
        addr = _resolve_name(name, socket.AF_INET6)
        QApplication.restoreOverrideCursor()
        if addr:
            self.address.setText(addr)
            self._on_changed()
            return
        QMessageBox.warning(
            self,
            self.tr('DNS Lookup Failed'),
            self.tr("DNS lookup failed for '%1'.").replace('%1', name),
        )


class NetworkDialog(_BaseAddressDialog):
    def __init__(self, parent=None):
        super().__init__('networkdialog_q.ui', parent)

    def _apply_changes(self):
        addr = self.address.text().strip()
        if addr and _validate_ipv4(addr) is None:
            QMessageBox.warning(
                self,
                self.tr('Invalid Address'),
                self.tr("Illegal IP address '%1'").replace('%1', addr),
            )
            return
        nm = self.netmask.text().strip()
        if nm and _validate_ipv4_netmask(nm) is None:
            QMessageBox.warning(
                self,
                self.tr('Invalid Netmask'),
                self.tr("Illegal netmask '%1'").replace('%1', nm),
            )
            return
        super()._apply_changes()

    @Slot()
    def addressEntered(self):
        """Parse CIDR notation (e.g. '192.168.1.0/24') into address + netmask."""
        text = self.address.text().strip()
        if '/' not in text:
            return
        try:
            net = ipaddress.IPv4Network(text, strict=False)
            self.address.setText(str(net.network_address))
            self.netmask.setText(str(net.netmask))
        except ValueError:
            pass


class NetworkDialogIPv6(_BaseAddressDialog):
    def __init__(self, parent=None):
        super().__init__('networkdialogipv6_q.ui', parent)

    def _apply_changes(self):
        addr = self.address.text().strip()
        if addr and _validate_ipv6(addr) is None:
            QMessageBox.warning(
                self,
                self.tr('Invalid Address'),
                self.tr("Illegal IPv6 address '%1'").replace('%1', addr),
            )
            return
        nm = self.netmask.text().strip()
        if nm:
            try:
                prefix = int(nm)
                if prefix < 1 or prefix > 127:
                    raise ValueError
            except ValueError:
                QMessageBox.warning(
                    self,
                    self.tr('Invalid Netmask'),
                    self.tr("Illegal netmask '%1'").replace('%1', nm),
                )
                return
        super()._apply_changes()

    @Slot()
    def addressEntered(self):
        """Parse CIDR notation (e.g. '2001:db8::/32') into address + prefix length."""
        text = self.address.text().strip()
        if '/' not in text:
            return
        try:
            net = ipaddress.IPv6Network(text, strict=False)
            self.address.setText(str(net.network_address))
            self.netmask.setText(str(net.prefixlen))
        except ValueError:
            pass


class AddressRangeDialog(BaseObjectDialog):
    def __init__(self, parent=None):
        super().__init__('addressrangedialog_q.ui', parent)

    def _populate(self):
        self.obj_name.setText(self._obj.name or '')
        start = self._obj.start_address or {}
        end = self._obj.end_address or {}
        self.rangeStart.setText(start.get('address', ''))
        self.rangeEnd.setText(end.get('address', ''))

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        start_text = self.rangeStart.text().strip()
        end_text = self.rangeEnd.text().strip()
        # Validate both addresses and ensure same family.
        start_addr = end_addr = None
        if start_text:
            try:
                start_addr = ipaddress.ip_address(start_text)
            except ValueError:
                QMessageBox.warning(
                    self,
                    self.tr('Invalid Address'),
                    self.tr("Illegal IP address '%1'").replace(
                        '%1',
                        start_text,
                    ),
                )
                return
        if end_text:
            try:
                end_addr = ipaddress.ip_address(end_text)
            except ValueError:
                QMessageBox.warning(
                    self,
                    self.tr('Invalid Address'),
                    self.tr("Illegal IP address '%1'").replace(
                        '%1',
                        end_text,
                    ),
                )
                return
        if start_addr and end_addr:
            if start_addr.version != end_addr.version:
                QMessageBox.warning(
                    self,
                    self.tr('Address Mismatch'),
                    self.tr(
                        "Range start '%1' and end '%2' must be the same IP version.",
                    )
                    .replace('%1', start_text)
                    .replace('%2', end_text),
                )
                return
            # Auto-correct: end must be >= start.
            if end_addr < start_addr:
                end_addr = start_addr
                self.rangeEnd.setText(str(end_addr))
        start = dict(self._obj.start_address or {})
        start['address'] = self.rangeStart.text().strip()
        self._obj.start_address = start
        end = dict(self._obj.end_address or {})
        end['address'] = self.rangeEnd.text().strip()
        self._obj.end_address = end

    @Slot()
    def addressEntered(self):
        """Adjust rangeEnd when rangeStart loses focus.

        If the start address contains a ``/``, parse it as CIDR and fill
        both start and end with the first/last host.  Otherwise, copy
        start to end when the end is empty or a different IP family.
        """
        text = self.rangeStart.text().strip()
        if not text:
            return
        try:
            if '/' in text:
                net = ipaddress.ip_network(text, strict=False)
                hosts = list(net.hosts())
                if hosts:
                    self.rangeStart.setText(str(hosts[0]))
                    self.rangeEnd.setText(str(hosts[-1]))
                else:
                    # Single-host network (e.g. /32 or /128).
                    self.rangeStart.setText(str(net.network_address))
                    self.rangeEnd.setText(str(net.network_address))
            else:
                start_addr = ipaddress.ip_address(text)
                end_text = self.rangeEnd.text().strip()
                if not end_text:
                    self.rangeEnd.setText(text)
                else:
                    end_addr = ipaddress.ip_address(end_text)
                    if end_addr.version != start_addr.version:
                        self.rangeEnd.setText(text)
        except ValueError:
            pass


class PhysAddressDialog(BaseObjectDialog):
    def __init__(self, parent=None):
        super().__init__('physaddressdialog_q.ui', parent)

    def _populate(self):
        self.obj_name.setText(self._obj.name or '')
        inet = self._obj.inet_addr_mask or {}
        self.pAddress.setText(inet.get('address', ''))

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        inet = dict(self._obj.inet_addr_mask or {})
        inet['address'] = self.pAddress.text()
        self._obj.inet_addr_mask = inet

    @Slot()
    def changed(self):
        pass  # https://github.com/Linuxfabrik/firewallfabrik/issues/14



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Editor panel dialog for time/interval objects."""

from PySide6.QtCore import QTime

from firewallfabrik.gui.base_object_dialog import BaseObjectDialog

# Map day-of-week index (0=Sun, fwbuilder convention) to the checkbox
# objectName in timedialog_q.ui.
_DOW_CHECKBOXES = {
    0: 'cbStart7_2',  # Sun
    1: 'cbStart1_2',  # Mon
    2: 'cbStart2_2',  # Tue
    3: 'cbStart3_2',  # Wed
    4: 'cbStart4_2',  # Thu
    5: 'cbStart5_2',  # Fri
    6: 'cbStart6_2',  # Sat
}


class TimeDialog(BaseObjectDialog):
    def __init__(self, parent=None):
        super().__init__('timedialog_q.ui', parent)

    def _populate(self):
        self.obj_name.setText(self._obj.name or '')
        data = self._obj.data or {}

        # Support both new format (from_time/to_time) and legacy format
        # (from_hour/from_minute/to_hour/to_minute) from .fwb imports.
        start_time = data.get('from_time', '')
        if not start_time:
            h = data.get('from_hour', '')
            m = data.get('from_minute', '')
            if h not in ('', '-1') and m not in ('', '-1'):
                start_time = f'{int(h):02d}:{int(m):02d}'

        end_time = data.get('to_time', '')
        if not end_time:
            h = data.get('to_hour', '')
            m = data.get('to_minute', '')
            if h not in ('', '-1') and m not in ('', '-1'):
                end_time = f'{int(h):02d}:{int(m):02d}'

        if start_time:
            self.startTime.setTime(QTime.fromString(start_time, 'HH:mm'))
        self.endTime.setTime(
            QTime.fromString(end_time, 'HH:mm') if end_time else QTime(23, 59)
        )

        days_str = data.get('days_of_week', '')
        active_days = set(days_str.split(',')) if days_str else set()
        for idx, cb_name in _DOW_CHECKBOXES.items():
            cb = getattr(self, cb_name, None)
            if cb:
                cb.setChecked(str(idx) in active_days)

    def _apply_changes(self):
        self._obj.name = self.obj_name.text()
        data = dict(self._obj.data or {})

        data['from_time'] = self.startTime.time().toString('HH:mm')
        data['to_time'] = self.endTime.time().toString('HH:mm')

        # Clean up legacy keys if present
        for key in (
            'from_date',
            'from_hour',
            'from_minute',
            'to_date',
            'to_hour',
            'to_minute',
        ):
            data.pop(key, None)

        active_days = []
        for idx, cb_name in _DOW_CHECKBOXES.items():
            cb = getattr(self, cb_name, None)
            if cb and cb.isChecked():
                active_days.append(str(idx))
        data['days_of_week'] = ','.join(active_days) if active_days else ''
        self._obj.data = data



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Database-mutating operations (CRUD) for the object tree."""

import copy
import uuid
from datetime import UTC, datetime

import sqlalchemy

from firewallfabrik.core.objects import (
    Address,
    Firewall,
    Group,
    Host,
    Interface,
    Interval,
    Library,
    Rule,
    RuleSet,
    Service,
    group_membership,
    rule_elements,
)
from firewallfabrik.gui.object_tree_data import (
    MODEL_MAP,
    NEW_TYPES_FOR_FOLDER,
    SYSTEM_GROUP_PATHS,
    find_group_by_path,
    normalize_subfolders,
)


def _stamp_parent_firewall(obj):
    """Update ``lastModified`` on the Firewall owning *obj* (if any).

    Returns the Firewall instance so the caller can emit tree updates,
    or *None* if *obj* is not under a Firewall.
    """
    fw = None
    if isinstance(obj, Firewall):
        fw = obj
    elif isinstance(obj, Interface):
        iface = obj
        while iface.parent_interface is not None:
            iface = iface.parent_interface
        device = iface.device
        if isinstance(device, Firewall):
            fw = device
    elif isinstance(obj, Address) and obj.interface is not None:
        return _stamp_parent_firewall(obj.interface)
    if fw is not None:
        data = dict(fw.data or {})
        data['lastModified'] = int(datetime.now(tz=UTC).timestamp())
        fw.data = data
    return fw


# All ORM classes that can own a ``library_id`` column.
_LIB_OWNED_CLASSES = (Address, Group, Host, Interface, Interval, Service)

# Ordered for safe cascade deletion (children before parents).
_ALL_ORM_CLASSES = (
    Address,
    Interval,
    Rule,
    Service,
    Interface,
    RuleSet,
    Host,
    Group,
    Library,
)


class TreeOperations:
    """Encapsulates all DB-mutating operations for the object tree."""

    def __init__(self, db_manager=None):
        self._db_manager = db_manager

    # ------------------------------------------------------------------
    # Delete — unified
    # ------------------------------------------------------------------

    @staticmethod
    def _collect_all_ids(session, root_id):
        """Recursively collect ALL descendant IDs from *root_id*.

        Handles: Host -> Interfaces -> Addresses, Host -> RuleSets -> Rules,
        Interface -> Addresses, Group -> members -> sub-groups.

        Returns ``(obj_ids: set, rule_ids: set)``.
        """
        obj_ids = {root_id}
        rule_ids = set()
        queue = [root_id]
        seen = {root_id}

        while queue:
            current_id = queue.pop()

            # Host -> interfaces + rule_sets
            host = session.get(Host, current_id)
            if host is not None:
                for iface in host.interfaces:
                    obj_ids.add(iface.id)
                    if iface.id not in seen:
                        seen.add(iface.id)
                        queue.append(iface.id)
                for rs in host.rule_sets:
                    obj_ids.add(rs.id)
                    for rule in rs.rules:
                        rule_ids.add(rule.id)
                        obj_ids.add(rule.id)

            # Interface -> addresses
            iface = session.get(Interface, current_id)
            if iface is not None:
                for addr in iface.addresses:
                    obj_ids.add(addr.id)

            # Group -> child objects + sub-groups
            group = session.get(Group, current_id)
            if group is not None:
                for addr in group.addresses:
                    obj_ids.add(addr.id)
                for svc in group.services:
                    obj_ids.add(svc.id)
                for itv in group.intervals:
                    obj_ids.add(itv.id)
                for dev in group.devices:
                    if dev.id not in seen:
                        seen.add(dev.id)
                        queue.append(dev.id)
                    obj_ids.add(dev.id)
                for child_grp in group.child_groups:
                    if child_grp.id not in seen:
                        seen.add(child_grp.id)
                        queue.append(child_grp.id)
                    obj_ids.add(child_grp.id)

        return obj_ids, rule_ids

    @staticmethod
    def _cleanup_references_and_delete(session, obj_ids, rule_ids):
        """Single-pass reference cleanup + cascade delete."""
        # 1. rule_elements by rule_id
        for rid in rule_ids:
            session.execute(
                rule_elements.delete().where(rule_elements.c.rule_id == rid)
            )

        # 2. rule_elements by target_id
        for oid in obj_ids:
            session.execute(
                rule_elements.delete().where(rule_elements.c.target_id == oid)
            )

        # 3. group_membership (both as member and as group)
        for oid in obj_ids:
            session.execute(
                group_membership.delete().where(group_membership.c.member_id == oid)
            )
            session.execute(
                group_membership.delete().where(group_membership.c.group_id == oid)
            )

        # 4. Delete all collected objects in dependency order (children
        #    before parents) to avoid cascade-nullify FK violations.
        #    _ALL_ORM_CLASSES is ordered: Address, Interval, Rule,
        #    Service, Interface, RuleSet, Host, Group, Library.
        #    Use no_autoflush to prevent premature flushing while
        #    session.get() loads objects.
        deleted = set()
        with session.no_autoflush:
            for cls in _ALL_ORM_CLASSES:
                for oid in obj_ids:
                    if oid in deleted:
                        continue
                    obj = session.get(cls, oid)
                    if obj is not None:
                        session.delete(obj)
                        deleted.add(oid)

    def delete_object(self, obj_id, model_cls, obj_name, obj_type, *, prefix=''):
        """Delete *obj_id* and clean up all references.  Returns True on success."""
        if self._db_manager is None:
            return False

        session = self._db_manager.create_session()
        try:
            obj = session.get(model_cls, obj_id)
            if obj is None:
                session.close()
                return False

            # Stamp the parent Firewall's lastModified *before* deleting
            # the child so the relationship is still traversable.
            _stamp_parent_firewall(obj)

            obj_ids, rule_ids = self._collect_all_ids(session, obj_id)
            self._cleanup_references_and_delete(session, obj_ids, rule_ids)

            session.commit()
            self._db_manager.save_state(f'{prefix}Delete {obj_type} {obj_name}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return True

    def delete_library(self, lib_id, lib_name):
        """Delete a library and all its contents.  Returns True on success."""
        if self._db_manager is None:
            return False

        session = self._db_manager.create_session()
        try:
            lib = session.get(Library, lib_id)
            if lib is None:
                return False

            # Collect IDs of ALL objects in this library.
            obj_ids = set()
            rule_ids = set()

            for cls in _LIB_OWNED_CLASSES:
                if not hasattr(cls, 'library_id'):
                    continue
                for obj in session.scalars(
                    sqlalchemy.select(cls).where(cls.library_id == lib_id)
                ).all():
                    sub_ids, sub_rules = self._collect_all_ids(session, obj.id)
                    obj_ids |= sub_ids
                    rule_ids |= sub_rules

            self._cleanup_references_and_delete(session, obj_ids, rule_ids)
            session.delete(lib)
            session.commit()
            self._db_manager.save_state(f'Delete Library {lib_name}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return True

    def delete_folder(self, lib_id, folder_name, *, parent_group_id=None):
        """Delete a user-created subfolder and all nested children.

        Moves all child objects to the parent level (removes their
        ``data.folder`` field) and removes the subfolder **and every
        nested child path** from the parent's ``data['subfolders']``.
        The parent is determined by *parent_group_id* (a Group) or
        falls back to the Library.
        """
        if self._db_manager is None:
            return

        child_prefix = folder_name + '/'

        def _folder_matches(folder):
            return folder == folder_name or folder.startswith(child_prefix)

        session = self._db_manager.create_session()
        try:
            if parent_group_id is not None:
                parent = session.get(Group, parent_group_id)
                if parent is not None:
                    for attr in (
                        'addresses',
                        'child_groups',
                        'devices',
                        'intervals',
                        'services',
                    ):
                        for child in getattr(parent, attr, []):
                            obj_data = child.data or {}
                            if _folder_matches(obj_data.get('folder', '')):
                                new_data = {
                                    k: v for k, v in obj_data.items() if k != 'folder'
                                }
                                child.data = new_data or None
            else:
                for cls in _LIB_OWNED_CLASSES:
                    if not hasattr(cls, 'data') or not hasattr(cls, 'library_id'):
                        continue
                    for obj in (
                        session.scalars(
                            sqlalchemy.select(cls).where(cls.library_id == lib_id)
                        )
                        .unique()
                        .all()
                    ):
                        obj_data = obj.data or {}
                        if _folder_matches(obj_data.get('folder', '')):
                            new_data = {
                                k: v for k, v in obj_data.items() if k != 'folder'
                            }
                            obj.data = new_data or None
                parent = session.get(Library, lib_id)

            # Remove folder and all nested child paths from subfolders list.
            if parent is not None:
                parent_data = copy.deepcopy(parent.data or {})
                subfolders = normalize_subfolders(parent_data.get('subfolders', []))
                subfolders = [s for s in subfolders if not _folder_matches(s)]
                if subfolders:
                    parent_data['subfolders'] = subfolders
                else:
                    parent_data.pop('subfolders', None)
                parent.data = parent_data

            session.commit()
            self._db_manager.save_state(f'Delete folder "{folder_name}"')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    # ------------------------------------------------------------------
    # Duplicate
    # ------------------------------------------------------------------

    def duplicate_object(
        self,
        source_id,
        model_cls,
        target_lib_id,
        *,
        folder=None,
        prefix='',
        target_group_id=None,
        target_interface_id=None,
    ):
        """Deep-copy *source_id* into *target_lib_id*. Returns new UUID or None.

        Optional *target_interface_id* / *target_group_id* place the clone
        under a specific interface or group instead of the library root.
        When *folder* is given, the clone's ``data.folder`` is set to that
        value (used when pasting into a user subfolder).
        """
        if self._db_manager is None:
            return None

        session = self._db_manager.create_session()
        try:
            source = session.get(model_cls, source_id)
            if source is None:
                session.close()
                return None

            id_map = {}
            new_obj = self._clone_object(source, id_map)

            # Clear all parent references first, then set the target.
            if hasattr(new_obj, 'interface_id'):
                new_obj.interface_id = None
            if hasattr(new_obj, 'group_id'):
                new_obj.group_id = None
            if hasattr(new_obj, 'parent_group_id'):
                new_obj.parent_group_id = None

            if target_interface_id is not None and hasattr(new_obj, 'interface_id'):
                new_obj.interface_id = target_interface_id
                # Addresses under interfaces don't carry library_id.
                if hasattr(new_obj, 'library_id'):
                    new_obj.library_id = None
            elif target_group_id is not None:
                if hasattr(new_obj, 'group_id'):
                    new_obj.group_id = target_group_id
                elif hasattr(new_obj, 'parent_group_id'):
                    new_obj.parent_group_id = target_group_id
                if hasattr(new_obj, 'library_id'):
                    new_obj.library_id = target_lib_id
            else:
                if hasattr(new_obj, 'library_id'):
                    new_obj.library_id = target_lib_id

            # Place clone in the target user subfolder, or clear folder
            # when pasting to the group root (empty string).
            if folder is not None and hasattr(new_obj, 'data'):
                data = copy.deepcopy(new_obj.data or {})
                if folder:
                    data['folder'] = folder
                else:
                    data.pop('folder', None)
                new_obj.data = data

            # Make name unique within the target scope.
            new_obj.name = self.make_name_unique(session, new_obj)

            session.add(new_obj)

            # Deep-copy children for devices (interfaces, rule sets, rules, rule elements).
            if isinstance(source, Host):
                self._duplicate_device_children(session, source, new_obj, id_map)

            # Copy group_membership entries for groups.
            if isinstance(source, Group):
                self._duplicate_group_members(session, source, new_obj)

            session.commit()
            self._db_manager.save_state(
                f'{prefix}Duplicate {model_cls.__name__} {source.name}',
            )
            new_id = new_obj.id
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return new_id

    @staticmethod
    def _clone_object(source, id_map):
        """Create a detached copy of *source* with a new UUID.

        All scalar/JSON column attributes are deep-copied.
        *id_map* is updated with ``{old_id: new_id}``.
        """
        mapper = sqlalchemy.inspect(type(source))
        new_id = uuid.uuid4()
        id_map[source.id] = new_id
        kwargs = {}
        for attr in mapper.column_attrs:
            key = attr.key
            if key == 'id':
                continue
            val = getattr(source, key)
            if isinstance(val, (dict, list, set)):
                val = copy.deepcopy(val)
            kwargs[key] = val
        return type(source)(id=new_id, **kwargs)

    def _duplicate_device_children(self, session, source_device, new_device, id_map):
        """Recursively duplicate interfaces, addresses, rule sets, rules, and rule elements."""
        # Interfaces + their child addresses.
        for iface in source_device.interfaces:
            new_iface = self._clone_object(iface, id_map)
            new_iface.device_id = new_device.id
            new_iface.library_id = new_device.library_id
            session.add(new_iface)
            for addr in iface.addresses:
                new_addr = self._clone_object(addr, id_map)
                new_addr.interface_id = new_iface.id
                new_addr.library_id = None
                new_addr.group_id = None
                session.add(new_addr)

        # Collect source rule_element rows and clone rule sets + rules.
        # We must read source rows before flushing the clones (the flush
        # would otherwise trigger lazy loads that might expire the source
        # relationships).
        rule_element_tasks = []
        for rs in source_device.rule_sets:
            new_rs = self._clone_object(rs, id_map)
            new_rs.device_id = new_device.id
            session.add(new_rs)
            for rule in rs.rules:
                new_rule = self._clone_object(rule, id_map)
                new_rule.rule_set_id = new_rs.id
                session.add(new_rule)
                rows = session.execute(
                    sqlalchemy.select(rule_elements).where(
                        rule_elements.c.rule_id == rule.id
                    )
                ).all()
                if rows:
                    rule_element_tasks.append((new_rule, rows))

        # Flush all ORM objects so that rule and rule_set rows exist in
        # the DB before we insert the raw rule_elements rows (which
        # reference them via FK).
        session.flush()

        for new_rule, rows in rule_element_tasks:
            for row in rows:
                target_id = id_map.get(row.target_id, row.target_id)
                session.execute(
                    rule_elements.insert().values(
                        rule_id=new_rule.id,
                        slot=row.slot,
                        target_id=target_id,
                        position=row.position,
                    )
                )

    @staticmethod
    def _duplicate_group_members(session, source_group, new_group):
        """Copy group_membership entries from *source_group* to *new_group*."""
        rows = session.execute(
            sqlalchemy.select(group_membership).where(
                group_membership.c.group_id == source_group.id
            )
        ).all()
        for row in rows:
            session.execute(
                group_membership.insert().values(
                    group_id=new_group.id,
                    member_id=row.member_id,
                    position=row.position,
                )
            )

    def duplicate_object_cross_db(
        self,
        source_db_manager,
        source_id,
        model_cls,
        target_lib_id,
        *,
        folder=None,
        prefix='',
        target_group_id=None,
        target_interface_id=None,
    ):
        """Deep-copy an object from *source_db_manager* into *this* database.

        Used for cross-file paste.  Reads the source object from the
        foreign database, serializes all scalar columns, and creates a
        new ORM instance in the local database with a fresh UUID.

        Returns the new object's UUID, or *None* on failure.
        """
        if self._db_manager is None or source_db_manager is None:
            return None

        # Read the source object from the foreign database.
        source_session = source_db_manager.create_session()
        target_session = self._db_manager.create_session()
        try:
            source = source_session.get(model_cls, source_id)
            if source is None:
                return None

            id_map = {}
            new_obj = self._clone_object(source, id_map)

            # Clear all parent references first, then set the target.
            if hasattr(new_obj, 'interface_id'):
                new_obj.interface_id = None
            if hasattr(new_obj, 'group_id'):
                new_obj.group_id = None
            if hasattr(new_obj, 'parent_group_id'):
                new_obj.parent_group_id = None

            if target_interface_id is not None and hasattr(new_obj, 'interface_id'):
                new_obj.interface_id = target_interface_id
                if hasattr(new_obj, 'library_id'):
                    new_obj.library_id = None
            elif target_group_id is not None:
                if hasattr(new_obj, 'group_id'):
                    new_obj.group_id = target_group_id
                elif hasattr(new_obj, 'parent_group_id'):
                    new_obj.parent_group_id = target_group_id
                if hasattr(new_obj, 'library_id'):
                    new_obj.library_id = target_lib_id
            else:
                if hasattr(new_obj, 'library_id'):
                    new_obj.library_id = target_lib_id

            if folder is not None and hasattr(new_obj, 'data'):
                data = copy.deepcopy(new_obj.data or {})
                if folder:
                    data['folder'] = folder
                else:
                    data.pop('folder', None)
                new_obj.data = data

            # Make name unique in the target database.
            new_obj.name = self.make_name_unique(target_session, new_obj)

            target_session.add(new_obj)

            # Deep-copy children for devices (interfaces, rule sets, etc.).
            if isinstance(source, Host):
                self._duplicate_device_children_cross_db(
                    source_session,
                    target_session,
                    source,
                    new_obj,
                    id_map,
                )

            # Copy group membership for groups.
            if isinstance(source, Group):
                self._duplicate_group_members_cross_db(
                    source_session,
                    target_session,
                    source,
                    new_obj,
                )

            target_session.commit()
            self._db_manager.save_state(
                f'{prefix}Paste {model_cls.__name__} {source.name} (cross-file)',
            )
            new_id = new_obj.id
        except Exception:
            target_session.rollback()
            raise
        finally:
            source_session.close()
            target_session.close()

        return new_id

    def _duplicate_device_children_cross_db(
        self,
        source_session,
        target_session,
        source_device,
        new_device,
        id_map,
    ):
        """Duplicate device children across databases.

        Like ``_duplicate_device_children`` but reads from *source_session*
        and writes to *target_session*.
        """
        for iface in source_device.interfaces:
            new_iface = self._clone_object(iface, id_map)
            new_iface.device_id = new_device.id
            new_iface.library_id = new_device.library_id
            target_session.add(new_iface)
            for addr in iface.addresses:
                new_addr = self._clone_object(addr, id_map)
                new_addr.interface_id = new_iface.id
                new_addr.library_id = None
                new_addr.group_id = None
                target_session.add(new_addr)

        rule_element_tasks = []
        for rs in source_device.rule_sets:
            new_rs = self._clone_object(rs, id_map)
            new_rs.device_id = new_device.id
            target_session.add(new_rs)
            for rule in rs.rules:
                new_rule = self._clone_object(rule, id_map)
                new_rule.rule_set_id = new_rs.id
                target_session.add(new_rule)
                rows = source_session.execute(
                    sqlalchemy.select(rule_elements).where(
                        rule_elements.c.rule_id == rule.id,
                    ),
                ).all()
                if rows:
                    rule_element_tasks.append((new_rule, rows))

        target_session.flush()

        for new_rule, rows in rule_element_tasks:
            for row in rows:
                # Remap target_id if it's a child of the cloned device;
                # otherwise keep the original (references to shared
                # Standard Library objects stay as-is — they won't
                # exist in the target DB but that's acceptable for
                # cross-file paste of rule elements referencing
                # non-standard objects; those will need manual fixup).
                target_id = id_map.get(row.target_id, row.target_id)
                target_session.execute(
                    rule_elements.insert().values(
                        rule_id=new_rule.id,
                        slot=row.slot,
                        target_id=target_id,
                        position=row.position,
                    ),
                )

    @staticmethod
    def _duplicate_group_members_cross_db(
        source_session,
        target_session,
        source_group,
        new_group,
    ):
        """Copy group_membership entries across databases.

        Only copies memberships whose member_id exists in the target DB.
        """
        rows = source_session.execute(
            sqlalchemy.select(group_membership).where(
                group_membership.c.group_id == source_group.id,
            ),
        ).all()
        for row in rows:
            # Check if the member exists in the target database before
            # inserting the membership (cross-file groups may reference
            # objects that don't exist in the target).
            exists = False
            for cls in (Address, Service, Interval, Host, Interface, Group):
                if target_session.get(cls, row.member_id) is not None:
                    exists = True
                    break
            if exists:
                target_session.execute(
                    group_membership.insert().values(
                        group_id=new_group.id,
                        member_id=row.member_id,
                        position=row.position,
                    ),
                )

    # ------------------------------------------------------------------
    # Move
    # ------------------------------------------------------------------

    def move_object(
        self,
        obj_id,
        model_cls,
        target_lib_id,
        *,
        folder=None,
        prefix='',
        target_group_id=None,
    ):
        """Move *obj_id* to *target_lib_id*. Returns True on success.

        When *target_group_id* and/or *folder* are given, the object is
        placed into that group / user subfolder (used for cut-paste into
        a subfolder).
        """
        if self._db_manager is None:
            return False

        session = self._db_manager.create_session()
        try:
            obj = session.get(model_cls, obj_id)
            if obj is None:
                session.close()
                return False

            obj_name = obj.name
            obj_type = getattr(obj, 'type', type(obj).__name__)
            obj.library_id = target_lib_id

            # Clear group/parent ownership — object lands at the library root.
            if hasattr(obj, 'group_id'):
                obj.group_id = None
            if hasattr(obj, 'parent_group_id'):
                obj.parent_group_id = None

            # Place into the target group when given.
            if target_group_id is not None:
                if hasattr(obj, 'group_id'):
                    obj.group_id = target_group_id
                elif hasattr(obj, 'parent_group_id'):
                    obj.parent_group_id = target_group_id

            # Set/clear user subfolder path.
            if folder is not None and hasattr(obj, 'data'):
                data = copy.deepcopy(obj.data or {})
                if folder:
                    data['folder'] = folder
                else:
                    data.pop('folder', None)
                obj.data = data

            # For devices, also move child interfaces.
            if isinstance(obj, Host):
                for iface in obj.interfaces:
                    iface.library_id = target_lib_id

            session.commit()
            self._db_manager.save_state(f'{prefix}Move {obj_type} {obj_name}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return True

    # ------------------------------------------------------------------
    # Create
    # ------------------------------------------------------------------

    def create_new_object(
        self,
        model_cls,
        type_name,
        lib_id,
        *,
        device_id=None,
        extra_data=None,
        folder=None,
        interface_id=None,
        name=None,
        parent_interface_id=None,
        prefix='',
    ):
        """Create a new object and return its UUID, or None on failure."""
        if self._db_manager is None:
            return None

        session = self._db_manager.create_session()
        try:
            new_id = uuid.uuid4()
            kwargs = {'id': new_id}

            # Only set 'type' for STI models that have a type column.
            if hasattr(model_cls, 'type'):
                kwargs['type'] = type_name

            # Library objects use database_id instead of library_id.
            if model_cls is Library:
                existing_lib = session.scalars(
                    sqlalchemy.select(Library).limit(1),
                ).first()
                if existing_lib is not None:
                    kwargs['database_id'] = existing_lib.database_id
                else:
                    session.close()
                    return None
            elif interface_id is not None and hasattr(model_cls, 'interface_id'):
                kwargs['interface_id'] = interface_id
            elif parent_interface_id is not None and hasattr(
                model_cls, 'parent_interface_id'
            ):
                kwargs['parent_interface_id'] = parent_interface_id
                if device_id is not None and hasattr(model_cls, 'device_id'):
                    kwargs['device_id'] = device_id
                if hasattr(model_cls, 'library_id'):
                    kwargs['library_id'] = lib_id
            elif device_id is not None and hasattr(model_cls, 'device_id'):
                kwargs['device_id'] = device_id
                if hasattr(model_cls, 'library_id'):
                    kwargs['library_id'] = lib_id
            else:
                if hasattr(model_cls, 'library_id'):
                    kwargs['library_id'] = lib_id

            # Use SYSTEM_GROUP_PATHS to resolve the correct (possibly
            # nested) group for this object type.  Falls back to the
            # virtual data.folder mechanism when the group doesn't exist.
            if folder and hasattr(model_cls, 'group_id') and 'group_id' not in kwargs:
                path = SYSTEM_GROUP_PATHS.get(type_name, '')
                target_group = find_group_by_path(session, lib_id, path)
                if target_group is not None:
                    kwargs['group_id'] = target_group.id
                    # System-level folder names (e.g. "Time") are only
                    # used for group resolution; user subfolder paths
                    # (e.g. "test1/test2") must be preserved in
                    # data.folder.
                    if folder in NEW_TYPES_FOR_FOLDER:
                        folder = None

            # Group-type objects use parent_group_id (not group_id) to
            # nest inside a folder group (e.g. ObjectGroup -> Objects/Groups).
            if (
                folder
                and hasattr(model_cls, 'parent_group_id')
                and 'parent_group_id' not in kwargs
            ):
                path = SYSTEM_GROUP_PATHS.get(type_name, '')
                target_group = find_group_by_path(session, lib_id, path)
                if target_group is not None:
                    kwargs['parent_group_id'] = target_group.id
                    if folder in NEW_TYPES_FOR_FOLDER:
                        folder = None

            # Extract options from extra_data (platform defaults for devices).
            options = extra_data.pop('options', None) if extra_data else None

            # Build data dict: merge folder and extra_data.
            data = {}
            if folder:
                data['folder'] = folder
            if extra_data:
                data.update(extra_data)
            if data and hasattr(model_cls, 'data'):
                kwargs['data'] = data

            if options and hasattr(model_cls, 'options'):
                kwargs['options'] = options

            new_obj = model_cls(**kwargs)
            new_obj.name = name or f'New {type_name}'

            # Make name unique.
            new_obj.name = self.make_name_unique(session, new_obj)

            # First RuleSet of a given type on a device is the top ruleset.
            if (
                model_cls is RuleSet
                and hasattr(new_obj, 'device_id')
                and new_obj.device_id is not None
            ):
                existing = session.scalar(
                    sqlalchemy.select(sqlalchemy.func.count())
                    .select_from(RuleSet)
                    .where(
                        RuleSet.device_id == new_obj.device_id,
                        RuleSet.type == type_name,
                    )
                )
                if not existing:
                    new_obj.top = True

            session.add(new_obj)
            session.commit()
            self._db_manager.save_state(f'{prefix}New {type_name}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return new_id

    def create_host_with_interfaces(
        self,
        lib_id,
        *,
        folder=None,
        interfaces=None,
        name=None,
        prefix='',
    ):
        """Create a Host with interfaces and addresses in one operation.

        Mirrors fwbuilder's ``newHostDialog::finishClicked()`` which
        creates the Host, its Interface children, and their IPv4/IPv6
        address children as a single undo-able action.

        Returns the new Host UUID, or None on failure.
        """
        if self._db_manager is None:
            return None

        session = self._db_manager.create_session()
        try:
            host_id = uuid.uuid4()
            host_kwargs = {
                'id': host_id,
                'type': 'Host',
                'library_id': lib_id,
            }
            # Place in the nested group (Objects/Hosts) if it exists.
            path = SYSTEM_GROUP_PATHS.get('Host', '')
            target_group = find_group_by_path(session, lib_id, path)
            if target_group is not None:
                host_kwargs['group_id'] = target_group.id
            elif folder:
                host_kwargs['data'] = {'folder': folder}
            host = Host(**host_kwargs)
            host.name = name or 'New Host'
            host.name = self.make_name_unique(session, host)
            session.add(host)

            for iface_data in interfaces or []:
                iface_id = uuid.uuid4()
                iface_name = iface_data.get('name', '')
                if not iface_name:
                    continue
                itype = iface_data.get('type', 0)
                iface = Interface(
                    id=iface_id,
                    device_id=host_id,
                    library_id=lib_id,
                    name=iface_name,
                    comment=iface_data.get('comment', ''),
                    data={
                        'dyn': str(itype == 1),
                        'label': iface_data.get('label', ''),
                        'security_level': '0',
                        'unnum': str(itype == 2),
                    },
                )
                session.add(iface)

                # Create IPv4/IPv6 address children (static interfaces only).
                if itype == 0:
                    for addr_info in iface_data.get('addresses', []):
                        addr_str = addr_info.get('address', '')
                        mask_str = addr_info.get('netmask', '')
                        is_v4 = addr_info.get('ipv4', True)
                        if not addr_str:
                            continue
                        addr_type = 'IPv4' if is_v4 else 'IPv6'
                        suffix = 'ip' if is_v4 else 'ip6'
                        addr_name = f'{host.name}:{iface_name}:{suffix}'
                        addr = Address(
                            id=uuid.uuid4(),
                            type=addr_type,
                            interface_id=iface_id,
                            name=addr_name,
                            inet_addr_mask={
                                'address': addr_str,
                                'netmask': mask_str,
                            },
                        )
                        session.add(addr)

            session.commit()
            self._db_manager.save_state(f'{prefix}New Host {host.name}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return host_id

    # ------------------------------------------------------------------
    # Folder move (drag & drop / cut-paste)
    # ------------------------------------------------------------------

    def set_object_folder(self, obj_id, model_cls, folder, *, prefix=''):
        """Set or clear ``data.folder`` on an existing object.

        *folder* is the target subfolder path (e.g. ``'A/B'``) or an
        empty string to move the object back to the group root.
        """
        if self._db_manager is None:
            return False

        session = self._db_manager.create_session()
        try:
            obj = session.get(model_cls, obj_id)
            if obj is None or not hasattr(obj, 'data'):
                session.close()
                return False
            data = copy.deepcopy(obj.data or {})
            old_folder = data.get('folder', '')
            if old_folder == folder:
                session.close()
                return False  # No-op.
            if folder:
                data['folder'] = folder
            else:
                data.pop('folder', None)
            obj.data = data
            session.commit()
            self._db_manager.save_state(
                f'{prefix}Move {getattr(obj, "type", type(obj).__name__)}'
                f' {obj.name} to folder'
            )
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    # ------------------------------------------------------------------
    # Subfolder & Rename
    # ------------------------------------------------------------------

    def create_subfolder(self, lib_id, name, *, parent_group_id=None):
        """Create a user subfolder in the parent's ``data.subfolders`` list.

        Matches fwbuilder's ``addSubfolderSlot()``: the subfolder is stored
        on whichever object the user right-clicked (a Group or a Library).
        """
        if self._db_manager is None:
            return

        session = self._db_manager.create_session()
        try:
            if parent_group_id is not None:
                parent = session.get(Group, parent_group_id)
            else:
                parent = session.get(Library, lib_id)
            if parent is None:
                return
            data = copy.deepcopy(parent.data or {})
            subfolders = normalize_subfolders(data.get('subfolders', []))
            if name in subfolders:
                return  # Already exists.
            subfolders.append(name)
            subfolders.sort(key=str.casefold)
            data['subfolders'] = subfolders
            parent.data = data
            session.commit()
            self._db_manager.save_state(f'New subfolder "{name}"')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def rename_folder(self, lib_id, old_name, new_name, *, parent_group_id=None):
        """Rename a category folder and update all nested child paths.

        Updates ``data.folder`` on every child object whose folder
        matches (or is nested under) *old_name*, and updates the
        parent's ``data['subfolders']`` list accordingly.
        """
        if self._db_manager is None:
            return

        old_prefix = old_name + '/'

        def _rename_folder_value(folder):
            if folder == old_name:
                return new_name
            if folder.startswith(old_prefix):
                return new_name + folder[len(old_name) :]
            return None

        session = self._db_manager.create_session()
        try:
            if parent_group_id is not None:
                parent_grp = session.get(Group, parent_group_id)
                if parent_grp is not None:
                    for attr in (
                        'addresses',
                        'child_groups',
                        'devices',
                        'intervals',
                        'services',
                    ):
                        for child in getattr(parent_grp, attr, []):
                            obj_data = child.data or {}
                            renamed = _rename_folder_value(obj_data.get('folder', ''))
                            if renamed is not None:
                                child.data = {**obj_data, 'folder': renamed}
                parent = parent_grp
            else:
                for cls in _LIB_OWNED_CLASSES:
                    if not hasattr(cls, 'data') or not hasattr(cls, 'library_id'):
                        continue
                    for obj in (
                        session.scalars(
                            sqlalchemy.select(cls).where(cls.library_id == lib_id)
                        )
                        .unique()
                        .all()
                    ):
                        obj_data = obj.data or {}
                        renamed = _rename_folder_value(obj_data.get('folder', ''))
                        if renamed is not None:
                            obj.data = {**obj_data, 'folder': renamed}
                parent = session.get(Library, lib_id)

            # Update parent's data['subfolders'] list — rename exact
            # match and all nested child paths.
            if parent is not None:
                parent_data = copy.deepcopy(parent.data or {})
                subfolders = normalize_subfolders(parent_data.get('subfolders', []))
                new_subfolders = []
                for s in subfolders:
                    renamed = _rename_folder_value(s)
                    new_subfolders.append(renamed if renamed is not None else s)
                new_subfolders.sort(key=str.casefold)
                parent_data['subfolders'] = new_subfolders
                parent.data = parent_data

            session.commit()
            self._db_manager.save_state(
                f'Rename folder \u201c{old_name}\u201d \u2192 \u201c{new_name}\u201d'
            )
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    # ------------------------------------------------------------------
    # Group
    # ------------------------------------------------------------------

    def group_objects(self, group_type, group_name, lib_id, member_ids):
        """Create a new group containing *member_ids*.  Returns new UUID or None."""
        if self._db_manager is None:
            return None

        # Determine the folder for this group type.
        folder = None
        for f, type_list in NEW_TYPES_FOR_FOLDER.items():
            if any(tn == group_type for tn, _dn in type_list):
                folder = f
                break

        new_id = self.create_new_object(
            MODEL_MAP[group_type],
            group_type,
            lib_id,
            folder=folder,
            name=group_name,
        )
        if new_id is None:
            return None

        # Add group_membership entries.
        session = self._db_manager.create_session()
        try:
            for pos, mid in enumerate(member_ids):
                session.execute(
                    group_membership.insert().values(
                        group_id=new_id,
                        member_id=uuid.UUID(mid) if isinstance(mid, str) else mid,
                        position=pos,
                    )
                )
            session.commit()
            self._db_manager.save_state(f'Group {len(member_ids)} objects')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return new_id

    # ------------------------------------------------------------------
    # Shared helpers
    # ------------------------------------------------------------------

    @staticmethod
    def make_name_unique(session, obj):
        """Return a unique name, appending '-1', '-2', ... only if needed.

        Queries the appropriate table for existing names.  If the base
        name is already free it is returned as-is (matching fwbuilder's
        ``makeNameUnique()``).
        """
        base_name = obj.name
        model_cls = type(obj)

        # Collect existing names in the same scope.
        stmt = sqlalchemy.select(model_cls.name).where(
            model_cls.name.like(f'{base_name}%')
        )
        if hasattr(obj, 'library_id') and obj.library_id is not None:
            stmt = stmt.where(model_cls.library_id == obj.library_id)
        existing = set(session.scalars(stmt).all())

        if base_name not in existing:
            return base_name

        suffix = 1
        while True:
            candidate = f'{base_name}-{suffix}'
            if candidate not in existing:
                return candidate
            suffix += 1

    # ------------------------------------------------------------------
    # Lock / Unlock
    # ------------------------------------------------------------------

    def lock_objects(self, obj_ids_with_types):
        """Set ``ro=True`` on the given objects.  Returns True on success."""
        if self._db_manager is None:
            return False
        session = self._db_manager.create_session()
        try:
            names = []
            for obj_id, obj_type in obj_ids_with_types:
                model_cls = MODEL_MAP.get(obj_type)
                if model_cls is None:
                    continue
                uid = uuid.UUID(obj_id) if isinstance(obj_id, str) else obj_id
                obj = session.get(model_cls, uid)
                if obj is not None and hasattr(obj, 'ro'):
                    obj.ro = True
                    names.append(obj.name)
            session.commit()
            label = ', '.join(names[:3]) + ('...' if len(names) > 3 else '')
            self._db_manager.save_state(f'Lock {label}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    def unlock_objects(self, obj_ids_with_types):
        """Set ``ro=False`` on the given objects.  Returns True on success."""
        if self._db_manager is None:
            return False
        session = self._db_manager.create_session()
        try:
            names = []
            for obj_id, obj_type in obj_ids_with_types:
                model_cls = MODEL_MAP.get(obj_type)
                if model_cls is None:
                    continue
                uid = uuid.UUID(obj_id) if isinstance(obj_id, str) else obj_id
                obj = session.get(model_cls, uid)
                if obj is not None and hasattr(obj, 'ro'):
                    obj.ro = False
                    names.append(obj.name)
            session.commit()
            label = ', '.join(names[:3]) + ('...' if len(names) > 3 else '')
            self._db_manager.save_state(f'Unlock {label}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    # ------------------------------------------------------------------
    # Keywords
    # ------------------------------------------------------------------

    def add_keyword(self, obj_ids_with_types, keyword):
        """Add *keyword* to the given objects.  Returns True on success."""
        if self._db_manager is None:
            return False
        session = self._db_manager.create_session()
        try:
            for obj_id, obj_type in obj_ids_with_types:
                model_cls = MODEL_MAP.get(obj_type)
                if model_cls is None:
                    continue
                uid = uuid.UUID(obj_id) if isinstance(obj_id, str) else obj_id
                obj = session.get(model_cls, uid)
                if obj is not None and hasattr(obj, 'keywords'):
                    kw = set(obj.keywords or set())
                    kw.add(keyword)
                    obj.keywords = kw
            session.commit()
            self._db_manager.save_state(f'Add keyword "{keyword}"')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    def remove_keyword(self, obj_ids_with_types, keyword):
        """Remove *keyword* from the given objects.  Returns True on success."""
        if self._db_manager is None:
            return False
        session = self._db_manager.create_session()
        try:
            for obj_id, obj_type in obj_ids_with_types:
                model_cls = MODEL_MAP.get(obj_type)
                if model_cls is None:
                    continue
                uid = uuid.UUID(obj_id) if isinstance(obj_id, str) else obj_id
                obj = session.get(model_cls, uid)
                if obj is not None and hasattr(obj, 'keywords'):
                    kw = set(obj.keywords or set())
                    kw.discard(keyword)
                    obj.keywords = kw
            session.commit()
            self._db_manager.save_state(f'Remove keyword "{keyword}"')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    # ------------------------------------------------------------------
    # Make subinterface
    # ------------------------------------------------------------------

    def make_subinterface(self, iface_id, target_parent_iface_id, *, prefix=''):
        """Reparent an interface under another interface.  Returns True on success."""
        if self._db_manager is None:
            return False
        session = self._db_manager.create_session()
        try:
            iface = session.get(Interface, iface_id)
            if iface is None:
                return False
            iface.parent_interface_id = target_parent_iface_id
            iface_name = iface.name
            session.commit()
            self._db_manager.save_state(f'{prefix}Make subinterface {iface_name}')
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
        return True

    # ------------------------------------------------------------------
    # Shared helpers
    # ------------------------------------------------------------------

    @staticmethod
    def get_all_tags(db_manager):
        """Collect every keyword used across all object tables."""
        if db_manager is None:
            return set()
        all_tags = set()
        with db_manager.session() as session:
            for cls in (Address, Group, Host, Interface, Interval, Service):
                for (tag_set,) in session.execute(sqlalchemy.select(cls.keywords)):
                    if tag_set:
                        all_tags.update(tag_set)
        return all_tags

    @staticmethod
    def get_writable_libraries(db_manager):
        """Return [(lib_id, lib_name), ...] for non-read-only libraries."""
        if db_manager is None:
            return []
        result = []
        with db_manager.session() as session:
            for lib in session.scalars(sqlalchemy.select(Library)).all():
                if not lib.ro:
                    result.append((lib.id, lib.name))
        result.sort(key=lambda t: t[1].lower())
        return result



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Comment editor panel for policy rules."""

from pathlib import Path

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QWidget

from firewallfabrik.gui.ui_loader import FWFUiLoader

_SAVE_DELAY_MS = 400


class CommentEditorPanel(QWidget):
    """Editor-pane widget for editing a policy rule's comment."""

    def __init__(self, parent=None):
        super().__init__(parent)
        ui_path = Path(__file__).resolve().parent / 'ui' / 'commenteditorpanel_q.ui'
        loader = FWFUiLoader(self)
        loader.load(str(ui_path))

        self._model = None
        self._index = None
        self._rule_id = None
        self._loading = False
        self._signals_connected = False

        # Debounce timer: save after a short pause to avoid a model
        # reload on every keystroke.
        self._save_timer = QTimer(self)
        self._save_timer.setSingleShot(True)
        self._save_timer.setInterval(_SAVE_DELAY_MS)
        self._save_timer.timeout.connect(self._save_comment)

    def load_rule(self, model, index):
        """Populate the editor from the rule at *index*."""
        self._save_timer.stop()
        self._disconnect_signals()
        self._model = model
        self._index = index
        row_data = model.get_row_data(index)
        self._rule_id = row_data.rule_id if row_data is not None else None
        self._load_comment()
        self._connect_signals()

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _load_comment(self):
        """Read the comment from the model and populate the editor."""
        self._loading = True
        try:
            editor = getattr(self, 'editor', None)
            if editor is None:
                return
            row_data = self._get_row_data()
            editor.setPlainText(row_data.comment if row_data is not None else '')
        finally:
            self._loading = False

    def _save_comment(self):
        """Persist the current editor text back to the model."""
        if self._model is None or self._index is None:
            return
        editor = getattr(self, 'editor', None)
        if editor is None:
            return
        new_comment = editor.toPlainText().strip()
        self._model.set_comment(self._index, new_comment)
        # set_comment() calls reload(), invalidating all QModelIndex objects.
        if self._rule_id is not None:
            self._index = self._model.index_for_rule(self._rule_id)

    def _on_text_changed(self):
        """Schedule a save after a short debounce delay."""
        if self._loading:
            return
        self._save_timer.start()

    def _get_row_data(self):
        """Return the row data for the current index."""
        if self._model is None or self._index is None:
            return None
        return self._model.get_row_data(self._index)

    # ------------------------------------------------------------------
    # Signal management
    # ------------------------------------------------------------------

    def _connect_signals(self):
        """Connect change signals to auto-save."""
        if self._signals_connected:
            return
        editor = getattr(self, 'editor', None)
        if editor is not None:
            editor.textChanged.connect(self._on_text_changed)
        self._signals_connected = True

    def _disconnect_signals(self):
        """Disconnect change signals to avoid stale callbacks."""
        if not self._signals_connected:
            return
        editor = getattr(self, 'editor', None)
        if editor is not None:
            editor.textChanged.disconnect(self._on_text_changed)
        self._signals_connected = False



# Copyright (C) 2026 Linuxfabrik <info@linuxfabrik.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# On Debian systems, the complete text of the GNU General Public License
# version 2 can be found in /usr/share/common-licenses/GPL-2.
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Base class for all object editor panel dialogs."""

from pathlib import Path

from PySide6.QtCore import Signal
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDateEdit,
    QDateTimeEdit,
    QLineEdit,
    QRadioButton,
    QSpinBox,
    QTimeEdit,
    QWidget,
)

from firewallfabrik.gui.comment_tags import CommentTags
from firewallfabrik.gui.ui_loader import FWFUiLoader

_UI_DIR = Path(__file__).resolve().parent / 'ui'


class BaseObjectDialog(QWidget):
    """Base class for all editor panel dialogs.

    Subclasses must implement ``_populate()`` and ``_apply_changes()``.
    """

    changed = Signal()

    def __init__(self, ui_filename, parent=None):
        super().__init__(parent)
        self._obj = None
        self._loading = False
        self._signals_connected = False
        loader = FWFUiLoader(self)
        loader.load(str(_UI_DIR / ui_filename))

    def load_object(self, obj, *, all_tags=None):
        """Load *obj* into the editor: populate widgets and connect signals."""
        self._obj = obj
        self._loading = True
        try:
            self._populate()
            comment_widget = self.findChild(CommentTags, 'commentKeywords')
            if comment_widget is not None:
                comment_widget.load(
                    getattr(obj, 'comment', None),
                    tags=getattr(obj, 'keywords', None),
                    all_tags=all_tags,
                )
            if not self._signals_connected:
                self._connect_change_signals()
                self._signals_connected = True
        finally:
            self._loading = False
        self._set_read_only(self._is_read_only())

    def _populate(self):
        """Fill widgets from ``self._obj``. Must be overridden."""
        raise NotImplementedError

    def _apply_changes(self):
        """Write widget values back to ``self._obj``. Must be overridden."""
        raise NotImplementedError

    def apply_all(self):
        """Apply subclass-specific changes *and* comment/keywords."""
        if self._is_read_only():
            return
        self._apply_changes()
        comment_widget = self.findChild(CommentTags, 'commentKeywords')
        if comment_widget is not None and self._obj is not None:
            self._obj.comment = comment_widget.get_comment()
            if hasattr(type(self._obj), 'keywords'):
                self._obj.keywords = comment_widget.get_tags()

    def _is_read_only(self):
        """Return ``True`` if the current object must not be modified."""
        obj = self._obj
        if obj is None:
            return False
        return getattr(obj, 'ro', False) or getattr(
            getattr(obj, 'library', None),
            'ro',
            False,
        )

    def _set_read_only(self, read_only):
        """Enable or disable all editable child widgets."""
        editable_types = (
            CommentTags,
            QCheckBox,
            QComboBox,
            QDateEdit,
            QDateTimeEdit,
            QLineEdit,
            QRadioButton,
            QSpinBox,
            QTimeEdit,
        )
        enabled = not read_only
        for child in self.findChildren(QWidget):
            if isinstance(child, editable_types):
                child.setEnabled(enabled)

    def _connect_change_signals(self):
        """Auto-connect child widget edit signals to ``_on_changed``."""
        for child in self.findChildren(QWidget):
            if isinstance(child, CommentTags):
                child.changed.connect(self._on_changed)
            elif isinstance(child, QLineEdit):
                child.editingFinished.connect(self._on_changed)
            elif isinstance(child, QCheckBox):
                child.stateChanged.connect(self._on_changed)
            elif isinstance(child, QComboBox):
                child.currentIndexChanged.connect(self._on_changed)
            elif isinstance(child, QSpinBox):
                child.valueChanged.connect(self._on_changed)
            elif isinstance(child, QRadioButton):
                child.toggled.connect(self._on_changed)
            elif isinstance(child, QDateEdit | QTimeEdit | QDateTimeEdit):
                child.dateTimeChanged.connect(self._on_changed)

    def _on_changed(self, *_args):
        if not self._loading:
            self.changed.emit()
